#!/usr/bin/env python

"""
Update identifiers for previous searches to version defined in config.

If a directory is passed, then the operation is performed on files in the directory.
If the path to a database file is passed, then the operation is performed on that
database.

Usage:

./update_identifiers.py path/to/output
./update_identifiers.py path/to/database.sqlite
"""
from sys import argv

import autofit as af
from autofit.tools.update_identifiers import update_directory_identifiers, update_database_identifiers

if __name__ == "__main__":
    argument = argv[1]

    if argument.endswith(
            ".sqlite"
    ):
        update_database_identifiers(
            af.database.open_database(
                argument
            )
        )
    else:
        update_directory_identifiers(
            argument
        )



import logging
from typing import Dict, Optional

import numpy as np
import os

from autofit.database.sqlalchemy_ import sa
from autofit.mapper.model_mapper import ModelMapper
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.fitness import Fitness
from autofit.non_linear.initializer import Initializer
from autofit.non_linear.search.mcmc.abstract_mcmc import AbstractMCMC
from autofit.non_linear.search.mcmc.auto_correlations import AutoCorrelationsSettings
from autofit.non_linear.search.mcmc.auto_correlations import AutoCorrelations
from autofit.non_linear.samples.sample import Sample
from autofit.non_linear.samples.mcmc import SamplesMCMC


class Zeus(AbstractMCMC):
    __identifier_fields__ = (
        "nwalkers",
        "tune",
        "tolerance",
        "patience",
        "mu",
        "light_mode",
    )

    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        unique_tag: Optional[str] = None,
        initializer: Optional[Initializer] = None,
        auto_correlation_settings=AutoCorrelationsSettings(),
        iterations_per_quick_update: int = None,
        iterations_per_full_update: int = None,
        number_of_cores: int = None,
        session: Optional[sa.orm.Session] = None,
        **kwargs
    ):
        """
        An Zeus non-linear search.

        For a full description of Zeus, checkout its Github and readthedocs webpages:

        https://github.com/minaskar/zeus

        https://zeus-mcmc.readthedocs.io/en/latest/

        If you use `Zeus` as part of a published work, please cite the package following the instructions under the
        *Attribution* section of the GitHub page.

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        nwalkers
            The number of walkers in the ensemble used to sample parameter space.
        nsteps
            The number of steps that must be taken by every walker. The `NonLinearSearch` will thus run for nwalkers *
            nsteps iterations.
        initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        auto_correlation_settings : AutoCorrelationsSettings
            Customizes and performs auto correlation calculations performed during and after the search.
        number_of_cores
            The number of cores Zeus sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        number_of_cores = number_of_cores or self._config("parallel", "number_of_cores")

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            initializer=initializer,
            auto_correlation_settings=auto_correlation_settings,
            iterations_per_quick_update=iterations_per_quick_update,
            iterations_per_full_update=iterations_per_full_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

        self.logger.debug("Creating Zeus Search")

    def _fit(self, model: AbstractPriorModel, analysis):
        """
        Fit a model using Zeus and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that inclues the maximum log likelihood instance and full
        chains used by the fit.
        """

        try:
            import zeus
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "\n--------------------\n"
                "You are attempting to perform a model-fit using Zeus. \n\n"
                "However, the optional library Zeus (https://zeus-mcmc.readthedocs.io/en/latest/) is "
                "not installed.\n\n"
                "Install it via the command `pip install zeus-mcmc==2.5.4`.\n\n"
                "----------------------"
            )

        pool = self.make_pool()

        fitness = Fitness(
            model=model,
            analysis=analysis,
            paths=self.paths,
            fom_is_log_likelihood=False,
            resample_figure_of_merit=-np.inf,
        )

        try:
            search_internal = self.paths.load_search_internal()

            state = search_internal.get_last_sample()
            log_posterior_list = search_internal.get_last_log_prob()

            samples = self.samples_from(model=model, search_internal=search_internal)

            total_iterations = search_internal.iteration

            if samples.converged:
                iterations_remaining = 0
            else:
                iterations_remaining = self.config_dict_run["nsteps"] - total_iterations

                self.logger.info(
                    "Resuming Zeus non-linear search (previous samples found)."
                )

        except (FileNotFoundError, AttributeError):
            search_internal = zeus.EnsembleSampler(
                nwalkers=self.config_dict_search["nwalkers"],
                ndim=model.prior_count,
                logprob_fn=fitness.call_wrap,
                pool=pool,
            )

            search_internal.ncall_total = 0

            (
                unit_parameter_lists,
                parameter_lists,
                log_posterior_list,
            ) = self.initializer.samples_from_model(
                total_points=search_internal.nwalkers,
                model=model,
                fitness=fitness,
                test_mode_samples=False,
                paths=self.paths,
                n_cores=self.number_of_cores,
            )

            self.plot_start_point(
                parameter_vector=parameter_lists[0],
                model=model,
                analysis=analysis,
            )

            state = np.zeros(shape=(search_internal.nwalkers, model.prior_count))

            self.logger.info(
                "Starting new Zeus non-linear search (no previous samples found)."
            )

            for index, parameters in enumerate(parameter_lists):
                state[index, :] = np.asarray(parameters)

            total_iterations = 0
            iterations_remaining = self.config_dict_run["nsteps"]

        while iterations_remaining > 0:
            if self.iterations_per_full_update > iterations_remaining:
                iterations = iterations_remaining
            else:
                iterations = self.iterations_per_full_update

            for sample in search_internal.sample(
                start=state,
                log_prob0=log_posterior_list,
                iterations=iterations,
                progress=True,
            ):
                pass

            search_internal.ncall_total += search_internal.ncall

            self.paths.save_search_internal(
                obj=search_internal,
            )

            state = search_internal.get_last_sample()
            log_posterior_list = search_internal.get_last_log_prob()

            total_iterations += iterations
            iterations_remaining = self.config_dict_run["nsteps"] - total_iterations

            samples = self.samples_from(model=model, search_internal=search_internal)

            if self.auto_correlation_settings.check_for_convergence:
                if (
                    search_internal.iteration
                    > self.auto_correlation_settings.check_size
                ):
                    if samples.converged:
                        iterations_remaining = 0

            auto_correlation_time = zeus.AutoCorrTime(
                samples=search_internal.get_chain()
            )

            discard = int(3.0 * np.max(auto_correlation_time))
            thin = int(np.max(auto_correlation_time) / 2.0)
            chain = search_internal.get_chain(discard=discard, thin=thin, flat=True)

            if "maxcall" in self.kwargs:
                if search_internal.ncall_total > self.kwargs["maxcall"]:
                    iterations_remaining = 0

            if iterations_remaining > 0:
                self.perform_update(
                    model=model,
                    analysis=analysis,
                    search_internal=search_internal,
                    fitness=fitness,
                    during_analysis=True,
                )

        return search_internal, fitness

    def samples_info_from(self, search_internal=None):
        search_internal = search_internal or self.paths.load_search_internal()

        auto_correlations = self.auto_correlations_from(search_internal=search_internal)

        return {
            "check_size": auto_correlations.check_size,
            "required_length": auto_correlations.required_length,
            "change_threshold": auto_correlations.change_threshold,
            "total_walkers": len(search_internal.get_chain()[0, :, 0]),
            "total_steps": int(search_internal.ncall_total),
            "time": self.timer.time if self.timer else None,
        }

    def samples_via_internal_from(self, model, search_internal=None):
        """
        Returns a `Samples` object from the zeus internal results.

        The samples contain all information on the parameter space sampling (e.g. the parameters,
        log likelihoods, etc.).

        The internal search results are converted from the native format used by the search to lists of values
        (e.g. `parameter_lists`, `log_likelihood_list`).

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        """

        search_internal = search_internal or self.paths.load_search_internal()

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1":

            samples_after_burn_in = search_internal.get_chain(
                discard=5, thin=5, flat=True
            )

        else:
            auto_correlations = self.auto_correlations_from(
                search_internal=search_internal
            )

            discard = int(3.0 * np.max(auto_correlations.times))
            thin = int(np.max(auto_correlations.times) / 2.0)
            samples_after_burn_in = search_internal.get_chain(
                discard=discard, thin=thin, flat=True
            )

            if len(samples_after_burn_in) == 0:

                logging.info(
                    """
                    After thinnng the Zeus samples in order to remove burn-in, no samples were left.
                    
                    To create a samples object containing samples, so that the code can continue and results
                    can be inspected, the full list of samples before removing burn-in has been used. This may 
                    indicate that the sampler has not converged and therefore your results may not be reliable.
                    
                    To fix this, run Zeus with more steps to ensure convergence is achieved or change the auto
                    correlation settings to be less aggressive in thinning samples.                
                    """
                )

                samples_after_burn_in = search_internal.get_chain(flat=True)

        parameter_lists = samples_after_burn_in.tolist()
        log_posterior_list = search_internal.get_log_prob(flat=True).tolist()
        log_prior_list = model.log_prior_list_from(parameter_lists=parameter_lists)

        log_likelihood_list = [
            log_posterior - log_prior
            for log_posterior, log_prior in zip(log_posterior_list, log_prior_list)
        ]

        weight_list = len(log_likelihood_list) * [1.0]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameter_lists,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list,
        )

        return SamplesMCMC(
            model=model,
            sample_list=sample_list,
            samples_info=self.samples_info_from(search_internal=search_internal),
            auto_correlation_settings=self.auto_correlation_settings,
            auto_correlations=self.auto_correlations_from(
                search_internal=search_internal
            ),
        )

    def auto_correlations_from(self, search_internal=None):
        import zeus

        search_internal = search_internal or self.paths.load_search_internal()

        times = zeus.AutoCorrTime(samples=search_internal.get_chain())
        try:
            previous_auto_correlation_times = zeus.AutoCorrTime(
                samples=search_internal.get_chain()[
                    : -self.auto_correlation_settings.check_size, :, :
                ],
            )
        except IndexError:
            self.logger.debug("Unable to compute previous auto correlation times.")
            previous_auto_correlation_times = None

        return AutoCorrelations(
            check_size=self.auto_correlation_settings.check_size,
            required_length=self.auto_correlation_settings.required_length,
            change_threshold=self.auto_correlation_settings.change_threshold,
            times=times,
            previous_times=previous_auto_correlation_times,
        )

    def config_dict_test_mode_from(self, config_dict: Dict) -> Dict:
        """
        Returns a configuration dictionary for test mode meaning that the sampler terminates as quickly as possible.

        Entries which set the total number of samples of the sampler (e.g. maximum calls, maximum likelihood
        evaluations) are reduced to low values meaning it terminates nearly immediately.

        Parameters
        ----------
        config_dict
            The original configuration dictionary for this sampler which includes entries controlling how fast the
            sampler terminates.

        Returns
        -------
        A configuration dictionary where settings which control the sampler's number of samples are reduced so it
        terminates as quickly as possible.
        """
        return {
            **config_dict,
            "nwalkers": 20,
            "nsteps": 10,
        }



from typing import Iterable


def add_notebook_quotes(
        lines: Iterable[str]
):
    """
    Add %% above and below docs quotes with triple quotes.

    Used for conversion to ipynb notebooks

    Parameters
    ----------
    lines
        An iterable of lines loaded from a notebook file

    Returns
    -------
    Lines with %% inserted before and after docs
    """
    out = list()
    is_in_quotes = False

    for line in lines[:-1]:
        if line.startswith('"""') or line.startswith("'''"):
            if is_in_quotes:
                out.extend(["'''", "\n\n", "# %%\n"])
            else:
                out.extend(["# %%", "\n", "'''\n"])

            is_in_quotes = not is_in_quotes
        else:
            out.append(line)

    return out


from typing import Dict, List

from autoconf import cached_property

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior.uniform import UniformPrior
from autofit.mapper.prior.gaussian import GaussianPrior
from autofit.mapper.prior.log_gaussian import LogGaussianPrior
from autofit.mapper.prior.log_uniform import LogUniformPrior
from autofit.mapper.prior.truncated_gaussian import TruncatedGaussianPrior
from autofit.mapper.prior_model.prior_model import ModelObject
from autofit.mapper.prior_model.prior_model import Model
from autofit.mapper.prior_model.collection import Collection


def str_for_object(obj: ModelObject) -> str:
    """
    Get a string representation of an object, including its id, type
    and important parameters. Used as labels for nodes.

    Parameters
    ----------
    obj
        Some component of the model.

    Returns
    -------
    A string representation of the object.
    """
    if isinstance(obj, Collection):
        return f"{obj.id}:Collection({len(obj)})"
    if isinstance(obj, Model):
        return f"{obj.id}:Model({obj.cls.__name__})"
    if isinstance(obj, UniformPrior):
        return f"{obj.id}:UniformPrior({obj.lower_limit}, {obj.upper_limit})"
    if isinstance(obj, GaussianPrior):
        return f"{obj.id}:GaussianPrior({obj.mean}, {obj.sigma})"
    if isinstance(obj, LogGaussianPrior):
        return f"{obj.id}:LogGaussianPrior({obj.mean}, {obj.sigma})"
    if isinstance(obj, LogUniformPrior):
        return f"{obj.id}:LogUniformPrior({obj.lower_limit}, {obj.upper_limit})"
    if isinstance(obj, TruncatedGaussianPrior):
        return f"{obj.id}:TruncatedGaussianPrior({obj.mean}, {obj.sigma}, ({obj.lower_limit}, {obj.upper_limit})"

    return repr(obj)


def generate_n_colors(n: int) -> List[str]:
    """
    Generate n distinct colors.
    """
    import colorsys

    colors = []
    for i in range(n):
        hue = i / n
        r, g, b = colorsys.hls_to_rgb(hue, 0.5, 1.0)
        hex_color = "#{:02x}{:02x}{:02x}".format(
            int(r * 255), int(g * 255), int(b * 255)
        )
        colors.append(hex_color)
    return colors


class VisualiseGraph:
    def __init__(self, model: AbstractPriorModel):
        """
        Visualise the graph of a model. Models and Priors are nodes
        and edges are the relationships between them.

        Parameters
        ----------
        model
            The model to visualise.
        """
        self.model = model

    def graph(self) -> "nx.DiGraph":
        """
        Generate a graph of the model with just edges. Could
        be superseded by the network method.
        """
        import networkx as nx

        graph = nx.DiGraph()

        def add_model(model):
            model_name = str_for_object(model)

            for name, prior in model.direct_prior_tuples:
                prior_name = str_for_object(prior)
                graph.add_edge(
                    model_name,
                    prior_name,
                    label=name,
                )

            for name, child_model in model.direct_prior_model_tuples:
                add_model(child_model)
                graph.add_edge(
                    model_name,
                    str_for_object(child_model),
                    label=name,
                )

        add_model(self.model)

        return graph

    @cached_property
    def colours(self) -> Dict[type, str]:
        """
        Generate a dictionary of colours for each type of object.
        """
        types = {
            Collection,
            UniformPrior,
            GaussianPrior,
            LogGaussianPrior,
            LogUniformPrior,
            TruncatedGaussianPrior,
        } | {model.cls for _, model in self.model.attribute_tuples_with_type(Model)}
        if isinstance(self.model, Model):
            types.add(self.model.cls)
        return {
            type_: color
            for type_, color in zip(
                sorted(types, key=str),
                generate_n_colors(len(types)),
            )
        }

    def network(self, notebook: bool = False) -> "Network":
        """
        Generate a network of the model including stylised nodes and edges.

        Models and Priors are nodes and edges are the relationships between them.

        Parameters
        ----------
        notebook
            Whether to display the network in a notebook. Must be true when calling
            the show method.

        Returns
        -------
        A network of the model.
        """
        from pyvis.network import Network

        net = Network(
            notebook=notebook,
            directed=True,
            cdn_resources="remote",
        )

        def add_model(obj, **kwargs):
            net.add_node(
                str_for_object(obj),
                shape="square",
                color=self.colours[obj.cls],
                size=15,
                **kwargs,
            )

        def add_collection(obj, **kwargs):
            net.add_node(
                str_for_object(obj),
                shape="hexagon",
                color=self.colours[Collection],
                size=15,
                **kwargs,
            )

        if isinstance(self.model, Model):
            add_model(self.model, borderWidth=5)
        else:
            add_collection(self.model, borderWidth=5)

        for _, model in self.model.attribute_tuples_with_type(Model):
            add_model(model)
        for _, collection in self.model.attribute_tuples_with_type(
            Collection,
            ignore_children=False,
        ):
            add_collection(collection)
        for _, prior in self.model.prior_tuples:
            net.add_node(
                str_for_object(prior),
                shape="dot",
                color=self.colours[type(prior)],
                size=10,
            )

        net.from_nx(self.graph())

        return net

    def save(self, path: str):
        """
        Save the network to a file. This should be an html file.

        Parameters
        ----------
        path
            The path to save the network to.
        """
        self.network().save_graph(path)

    def show(self, name: str):
        """
        Show the network.
        """
        self.network(notebook=True).show(f"{name}.html")



from typing import Set, Optional, Type, List, Tuple, Dict

from autofit.mapper.model import ModelInstance
from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior_model.collection import Collection
from autofit.mapper.prior_model.prior_model import Model
from autofit.mapper.variable import Plate
from autofit.messages import NormalMessage
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.tools.namer import namer

from .abstract import AbstractModelFactor


class HierarchicalFactor(Model):
    _plates: Tuple[Plate, ...] = ()

    def __init__(
        self,
        distribution: Type[Prior],
        optimiser=None,
        name: Optional[str] = None,
        use_jax : bool = False,
        **kwargs,
    ):
        """
        Associates variables in the graph with a distribution. That is,
        the optimisation prefers instances of the variables which better
        match the distribution. Both the distribution and variables sampled
        from it are optimised during a factor optimisation. This allows
        expectations from other factors to influence the optimisation of
        the distribution and vice-versa.

        Each HierarchicalFactor actually generates multiple factors - one
        for each associated variables - as this avoids optimisation of a
        high dimensionality factor.

        Question: would it make sense to experiment with a hierarchical
        factor that optimises all variables samples from a distribution
        simultaneously?

        Parameters
        ----------
        distribution
            A distribution from which variables are drawn
        optimiser
            An optional optimiser for this factor
        name
            An optional name for this factor
        kwargs
            Constants or Priors passed to the distribution to parameterize
            it. For example, a GaussianPrior requires mean and sigma arguments

        Examples
        --------
        factor = g.HierarchicalFactor(
            af.GaussianPrior,
            mean=af.GaussianPrior(
                mean=100,
                sigma=10
            ),
            sigma=af.GaussianPrior(
                mean=10,
                sigma=5
            )
        )
        factor.add_sampled_variable(
            prior
        )
        """
        super().__init__(distribution, **kwargs)
        self._name = name or namer(self.__class__.__name__)
        self._factors = list()
        self.optimiser = optimiser
        self._use_jax = use_jax

    @property
    def name(self):
        return self._name

    @property
    def prior_model(self):
        return self

    @property
    def plates(self):
        return self._plates

    def add_drawn_variable(self, prior: Prior):
        """
        Add a variable which is drawn from the distribution. This
        is likely the attribute of a FactorModel in the graph.

        Parameters
        ----------
        prior
            A variable which is sampled from the distribution.
        """
        self._factors.append(_HierarchicalFactor(self, prior))

    @property
    def factors(self) -> List["_HierarchicalFactor"]:
        """
        One factor is generated for each variable sampled from the
        distribution.
        """
        return self._factors


class Factor:
    def __init__(self, distribution_model: HierarchicalFactor):
        """
        Provides the function called by the factor in the graph.

        Parameters
        ----------
        distribution_model
            A prior model which parameterizes a distribution from which it
            is assumed the variable is drawn
        """
        self.distribution_model = distribution_model

    def __call__(self, **kwargs):
        """
        Call the factor

        Parameters
        ----------
        kwargs
            Maps names containing prior ids as well as a name 'argument' to values
            for each priors and the argument with which the distribution is called.

        Returns
        -------
        The probability that the argument was drawn from the distribution
        """
        argument = kwargs.pop("argument")
        arguments = dict()
        for name_, array in kwargs.items():
            prior_id = int(name_.split("_")[1])
            prior = self.distribution_model.prior_with_id(prior_id)
            arguments[prior] = array
        return self.distribution_model.instance_for_arguments(arguments).message(
            argument
        )


class _HierarchicalFactor(AbstractModelFactor):
    def __init__(
        self, distribution_model: HierarchicalFactor, drawn_prior: Prior,
    ):
        """
        A factor that links a variable to a parameterised distribution.

        Parameters
        ----------
        distribution_model
            A prior model which parameterizes a distribution from which it
            is assumed the variable is drawn
        drawn_prior
            A prior representing a variable which was drawn from the distribution
        """
        self.distribution_model = distribution_model
        self.drawn_prior = drawn_prior
        self._use_jax = distribution_model._use_jax

        prior_variable_dict = {prior.name: prior for prior in distribution_model.priors}

        prior_variable_dict["argument"] = drawn_prior

        super().__init__(
            prior_model=Collection(
                distribution_model=distribution_model, drawn_prior=drawn_prior
            ),
            factor=Factor(distribution_model),
            optimiser=distribution_model.optimiser,
            prior_variable_dict=prior_variable_dict,
            name=distribution_model.name,
        )

    @property
    def message_dict(self) -> Dict[Prior, NormalMessage]:
        """
        Dictionary mapping priors to messages. Does not account for inverse cavity
        behaviour as this caused bugs for hierarchical factors.
        """
        return {prior: prior.message for prior in self.priors}

    @property
    def variable(self):
        return self.drawn_prior

    def log_likelihood_function(self, instance):
        return instance.distribution_model.message(instance.drawn_prior, xp=self._xp)

    @property
    def priors(self) -> Set[Prior]:
        """
        The set of priors associated with this factor. This is the priors used
        to parameterize the distribution, plus an additional prior for the
        variable drawn from the distribution.
        """
        priors = super().priors
        priors.add(self.drawn_prior)
        return priors

    @property
    def analysis(self):
        return self

    def visualize(
        self, paths: AbstractPaths, instance: ModelInstance, during_analysis: bool
    ):
        pass



import math
import numpy as np
from typing import Tuple

"""
The `Gaussian` class in this module is the model components that is fitted to data using a non-linear search. The
inputs of its __init__ constructor are the parameters which can be fitted for.

The log_likelihood_function in the Analysis class receives an instance of this classes where the values of its
parameters have been set up according to the non-linear search. Because instances of the classes are used, this means
their methods (e.g. model_data_from) can be used in the log likelihood function.
"""


class Gaussian:
    def __init__(
        self,
        centre: float = 0.0,  # <- PyAutoFit recognises these constructor arguments
        normalization: float = 0.1,  # <- are the Gaussian`s model parameters.
        sigma: float = 0.01,
    ):
        """
        Represents a 1D `Gaussian` profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a non-linear search.

        Parameters
        ----------
        centre
            The x coordinate of the profile centre.
        normalization
            Overall normalization normalisation of the `Gaussian` profile.
        sigma
            The sigma value controlling the size of the Gaussian.
        """
        self.centre = centre
        self.normalization = normalization
        self.sigma = sigma

    @property
    def fwhm(self) -> float:
        """
        The full-width half-maximum of the Gaussian profile.

        This is used to illustrate latent variables in **PyAutoFit**, which are values that can be inferred from
        the free parameters of the model which we are interested and may want to store the full samples information
        on (e.g. to create posteriors).
        """
        return 2 * np.sqrt(2 * np.log(2)) * self.sigma

    def _tree_flatten(self):
        return (self.centre, self.normalization, self.sigma), None

    @classmethod
    def _tree_unflatten(cls, aux_data, children):
        return Gaussian(*children)

    def __eq__(self, other):
        return (
            isinstance(other, Gaussian)
            and self.centre == other.centre
            and self.normalization == other.normalization
            and self.sigma == other.sigma
        )

    def model_data_from(self, xvalues: np.ndarray, xp=np) -> np.ndarray:
        """
        Calculate the normalization of the profile on a 1D grid of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Gaussian, using its centre.

        Parameters
        ----------
        xvalues
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = xvalues - self.centre

        return xp.multiply(
            xp.divide(self.normalization, self.sigma * xp.sqrt(2.0 * xp.pi)),
            xp.exp(-0.5 * xp.square(xp.divide(transformed_xvalues, self.sigma))),
        )

    def f(self, x: float, xp=np):
        return (
            self.normalization
            / (self.sigma * xp.sqrt(2 * math.pi))
            * xp.exp(-0.5 * ((x - self.centre) / self.sigma) ** 2)
        )

    def __call__(self, xvalues: np.ndarray) -> np.ndarray:
        """
        For certain graphical models, the `__call__` function is overwritten for producing the model-fit.
        We include this here so these examples work, but it should not be important for most PyAutoFit users.

        Parameters
        ----------
        xvalues
            The x coordinates in the original reference frame of the grid.
        """
        return self.model_data_from(xvalues=xvalues)

    def inverse(self, y):
        """
        For graphical models, the inverse of the Gaussian is used to test certain aspects of the calculation.
        """

        a = self.normalization / (y * self.sigma * math.sqrt(2 * math.pi))

        b = 2 * math.log(a)

        return self.centre + self.sigma * math.sqrt(b)


class Exponential:
    def __init__(
        self,
        centre: float = 0.0,  # <- PyAutoFit recognises these constructor arguments are the model
        normalization: float = 0.1,  # <- parameters of the Gaussian.
        rate: float = 0.01,
    ):
        """
        Represents a 1D Exponential profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a `NonLinearSearch`.

        Parameters
        ----------
        centre
            The x coordinate of the profile centre.
        normalization
            Overall normalization normalisation of the `Gaussian` profile.
        rate
            The decay rate controlling has fast the Exponential declines.
        """
        self.centre = centre
        self.normalization = normalization
        self.rate = rate

    def model_data_from(self, xvalues: np.ndarray, xp=np) -> np.ndarray:
        """
        Calculate the 1D Gaussian profile on a 1D grid of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Exponential, using its centre.

        Parameters
        ----------
        values
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = xp.subtract(xvalues, self.centre)
        return self.normalization * xp.multiply(
            self.rate, xp.exp(-1.0 * self.rate * abs(transformed_xvalues))
        )

    def __call__(self, xvalues: np.ndarray) -> np.ndarray:
        """
        Calculate the 1D Gaussian profile on a 1D grid of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Exponential, using its centre.

        Parameters
        ----------
        values
            The x coordinates in the original reference frame of the grid.
        """
        return self.model_data_from(xvalues=xvalues)


class PhysicalNFW:
    def __init__(
        self,
        centre: Tuple[float, float],
        ell_comps: Tuple[float, float],
        log10m: float,
        concentration: float,
    ):
        self.centre = centre
        self.ell_comps = ell_comps
        self.log10m = log10m
        self.concentration = concentration



from collections import defaultdict
from typing import List


def find_groups(path_value_tuples, limit=1):
    """
    Groups path-value tuples by their path, up to a limit. This is used to group priors in the model representation.

    If multiple paths share the same suffix they are grouped together.

    Parameters
    ----------
    path_value_tuples
        A list of tuples of paths and values in the model
    limit
        How far from the end of paths grouping should terminate. This can be used to prevent grouping of
        priors and attributes.

    Returns
    -------
    A list of tuples of paths and values, where paths are grouped by their suffix.
    """
    try:
        maximum_path_length = max(len(path) for path, _ in path_value_tuples)
    except ValueError:
        return path_value_tuples
    for position in range(maximum_path_length - limit):
        path_value_tuples = _find_groups(path_value_tuples, position)
    return path_value_tuples


def integers_representative_key(integers: List[int]) -> str:
    """
    Given a list of integers, return a string that represents them in a concise way.

    e.g.
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] -> "1 - 10"
    [1, 2, 3, 5, 6, 7, 8, 9, 10] -> "1 - 3, 5 - 10"
    [1, 2, 3, 5, 6, 7, 8, 9, 10, 12] -> "1 - 3, 5 - 10, 12"

    Parameters
    ----------
    integers
        The list of integers to represent

    Returns
    -------
    A string representing the integers in a concise way
    """
    integers = sorted(integers)
    ranges = []
    start = integers[0]
    end = integers[0]

    for integer in integers[1:]:
        if integer == end + 1:
            end = integer
        else:
            ranges.append((start, end))
            start = integer
            end = integer

    ranges.append((start, end))

    return ", ".join(
        f"{start} - {end}" if start != end else str(start) for start, end in ranges
    )


def _find_groups(path_value_tuples, position):
    groups = defaultdict(list)
    paths = []
    for path, value in path_value_tuples:
        try:
            root_name = path[position]
            before = path[:position]
            after = path[position + 1 :]
            groups[(tuple(before), tuple(after), value)].append(root_name)
        except IndexError:
            paths.append((path, value))

    for (before, after, value), names in groups.items():
        try:
            names = list(map(int, names))
            representative_key = integers_representative_key(names)
        except ValueError:
            representative_key = (
                f"{min(names)} - {max(names)}" if len(set(names)) > 1 else names[0]
            )

        paths.append(((*before, representative_key, *after), value))
    return paths



import logging
from pathlib import Path
from typing import Optional, Union

from .. import model as m
from ..sqlalchemy_ import sa
from autofit.aggregator.search_output import SearchOutput

logger = logging.getLogger(__name__)


class Scraper:
    def __init__(
        self,
        directory: Union[Path, str],
        session: sa.orm.Session,
        reference: Optional[dict] = None,
        completed_only: bool = True,
    ):
        """
        Facilitates scraping of data output into a directory
        into the database.

        Parameters
        ----------
        directory
            A directory in which data has been stored
        session
            A database session
        reference
            A dictionary mapping search names to model paths
        completed_only
            If True, only completed fits will be scraped
        """
        self.directory = directory
        self.session = session
        self.reference = reference

        from autofit.aggregator.aggregator import Aggregator as ClassicAggregator

        self.aggregator = ClassicAggregator.from_directory(
            self.directory,
            reference=self.reference,
            completed_only=completed_only,
        )

    def scrape(self):
        """
        Recursively scrape fits from the directory and
        add them to the session
        """
        for fit in self._fits():
            self.session.add(fit)
        for grid_search in self._grid_searches():
            self.session.add(grid_search)

    def _fits(self):
        """
        Scrape data output into a directory tree so it can be added to the
        aggregator database.

        Returns
        -------
        Generator yielding Fit database objects
        """
        logger.info(f"Scraping directory {self.directory}")
        logger.info(f"{len(self.aggregator)} searches found")
        for item in self.aggregator:
            logger.info(
                f"Creating fit for: "
                f"{item.path_prefix} "
                f"{item.unique_tag} "
                f"{item.name} "
                f"{item.id} "
            )

            try:
                fit = self._retrieve_model_fit(item)
                logger.warning(f"Fit already existed with identifier {item.id}")
            except sa.orm.exc.NoResultFound:
                fit = m.Fit(
                    id=item.id,
                    name=item.name,
                    unique_tag=item.unique_tag,
                    model=item.model,
                    instance=item.instance,
                    is_complete=item.is_complete,
                    info=item.info,
                    max_log_likelihood=item.max_log_likelihood,
                    parent_id=item.parent_identifier,
                )

            _add_files_fit(fit, item)
            for i, child_analysis in enumerate(item.children):
                child_fit = m.Fit(
                    id=f"{item.id}_{i}",
                )
                _add_files_fit(child_fit, child_analysis)
                fit.children.append(child_fit)

            yield fit

    def _grid_searches(self):
        """
        Retrieve grid searches recursively from an output directory by
        searching for the .is_grid_search file.

        Should be called after adding Fits as it relies on querying fits

        Yields
        ------
        Fit objects representing grid searches with child fits associated
        """
        for item in self.aggregator.grid_searches():
            grid_search = m.Fit(
                id=item.id,
                unique_tag=item.unique_tag,
                is_grid_search=True,
                is_complete=item.is_complete,
            )

            _add_files(grid_search, item)

            for search in item.children:
                fit = self._retrieve_model_fit(search)
                grid_search.children.append(fit)
            yield grid_search

    def _retrieve_model_fit(self, item) -> m.Fit:
        """
        Retrieve a Fit, if one exists, corresponding to a given SearchOutput

        Parameters
        ----------
        item
            A SearchOutput from the classic Aggregator

        Returns
        -------
        A fit with the corresponding identifier

        Raises
        ------
        NoResultFound
            If no fit is found with the identifier
        """
        return self.session.query(m.Fit).filter(m.Fit.id == item.id).one()


def _add_files_fit(fit: m.Fit, item: SearchOutput):
    try:
        fit.samples = item.samples
    except AttributeError:
        logger.warning(f"Failed to load samples for {fit.id}")
    try:
        fit.latent_samples = item.latent_samples
    except AttributeError:
        logger.warning(f"Failed to load latent variables for {fit.id}")

    _add_files(fit, item)


def _add_files(fit: m.Fit, item: SearchOutput):
    """
    Load files from the path and add them to the database.

    Parameters
    ----------
    fit
        A fit to which the pickles belong
    """
    for json_output in item.jsons:
        fit.set_json(json_output.name, json_output.dict)

    for pickle_output in item.pickles:
        fit.set_pickle(pickle_output.name, pickle_output.value)

    for array_output in item.arrays:
        if array_output.name in ("samples", "latent_samples"):
            continue
        try:
            fit.set_array(array_output.name, array_output.value)
        except ValueError:
            logger.debug(f"Failed to load array {array_output.name} for {fit.id}")

    for fits in item.fits:
        fit.set_fits(fits.name, fits.value)



import inspect
from collections.abc import Iterable
from hashlib import md5
from typing import Optional

from autoconf.class_path import get_class_path

# floats are rounded to this increment so floating point errors
# have no impact on identifier value
RESOLUTION = 1e-8


class IdentifierField:
    """
    A field that the identifier depends on.

    If the value of the field is changed then the identifier
    must be recomputed prior to use.
    """

    def __set_name__(self, owner: object, name: str):
        """
        Called on instantiation

        Parameters
        ----------
        owner
            The object for which this field is an attribute
        name
            The name of the attribute
        """
        self.private = f"_{name}"
        setattr(owner, self.private, None)

    def __get__(self, obj: object, objtype=None) -> Optional:
        """
        Retrieve the value of this field.

        Parameters
        ----------
        obj
            The object for which this field is an attribute
        objtype

        Returns
        -------
        The value (or None if it has not been set)
        """
        return getattr(obj, self.private)

    def __set__(self, obj, value):
        """
        Set a value for this field

        Parameters
        ----------
        obj
            The object for which the field is an attribute
        value
            A new value for the attribute
        """
        obj._identifier = None
        setattr(obj, self.private, value)


class Identifier:
    def __init__(self, obj):
        """
        Wraps an object and recursively generates an identifier

        The version can be set in general.ini (output/identifier_version). It can
        be overridden by specifying it explicitly in the constructor.
        """
        self.hash_list = list()
        self._add_value_to_hash_list(obj)

    @property
    def description(self):
        return "\n".join(self.hash_list)

    def _add_value_to_hash_list(self, value):
        """
        Add some object and recursively add its children to the hash_list.

        The md5 hash of this object is taken to create an identifier.

        If an object specifies __identifier_fields__ then only attributes
        with a name in this list are included.

        Parameters
        ----------
        value
            An object
        """
        from .model_object import ModelObject
        from autofit.mapper.prior.constant import Constant

        if isinstance(value, Constant):
            self.add_value_to_hash_list(value.value)
            return

        if inspect.isclass(value):
            self.add_value_to_hash_list(get_class_path(value))
            return

        if isinstance(value, Exception):
            raise value
        if hasattr(value, "__dict__"):
            if hasattr(value, "__class__"):
                self.add_value_to_hash_list(value.__class__.__name__)
            d = value.__dict__

            if hasattr(value, "__identifier_fields__"):
                fields = value.__identifier_fields__

                try:
                    d = {k: getattr(value, k) for k in fields}
                except AttributeError as e:
                    raise AssertionError(
                        f"Missing identifier fields for {type(value)}"
                    ) from e
            elif (
                hasattr(value, "__class__")
                and not inspect.isclass(value)
                and not isinstance(value, ModelObject)
            ):
                args = inspect.getfullargspec(value.__class__).args
                d = {k: v for k, v in d.items() if k in args}
                if hasattr(value, "__exclude_identifier_fields__"):
                    excluded_fields = value.__exclude_identifier_fields__

                    d = {k: v for k, v in d.items() if k not in excluded_fields}
            self.add_value_to_hash_list(d)
        elif isinstance(value, dict):
            for key, value in value.items():
                key = str(key)
                if not (key.startswith("_") or key in ("id", "paths")):
                    self.hash_list.append(key)
                    self.add_value_to_hash_list(value)
        elif isinstance(value, float):
            try:
                value = RESOLUTION * round(value / RESOLUTION)
            except OverflowError:
                pass

            self.hash_list.append(str(value))
        elif isinstance(value, (str, int, bool)):
            self.hash_list.append(str(value))
        elif isinstance(value, Iterable):
            for value in value:
                self.add_value_to_hash_list(value)

    def add_value_to_hash_list(self, value):
        if isinstance(value, property):
            return
        self._add_value_to_hash_list(value)

    def __str__(self):
        return md5(".".join(self.hash_list).encode("utf-8")).hexdigest()

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"

    def __eq__(self, other):
        return str(self) == str(other)



from typing import List
from autofit.mapper.model import ModelInstance


class InterpolatorPath:
    def __init__(self, keys: List[str]):
        """
        Addresses a given attribute in a ModelInstance

        Parameters
        ----------
        keys
            A list of attribute names
        """
        self.keys = keys

    def __hash__(self):
        return hash(tuple(self.keys))

    def __repr__(self):
        return f"InterpolatorPath({self.keys})"

    def __getattr__(self, item: str) -> "InterpolatorPath":
        """
        Add a new attribute name to the end of the path
        """
        return InterpolatorPath(self.keys + [item])

    def get_value(self, instance: ModelInstance) -> float:
        """
        Retrieve the value at the path for a given instance.

        Parameters
        ----------
        instance
            An instance of some model

        Returns
        -------
        The value for the instance at the given path.
        """
        for key in self.keys:
            instance = getattr(instance, key)
        return instance

    def __eq__(self, other: float) -> "Equality":
        """
        Create an object describing the value the addressed attribute
        should have for interpolation.

        Parameters
        ----------
        other
            A value to which the instance will be interpolated

        Returns
        -------
        An object describing how to interpolate
        """
        return Equality(self, other)


class Equality:
    def __init__(self, path: InterpolatorPath, value: float):
        """
        Describes the value of a given attribute for which other values
        are interpolated.

        Parameters
        ----------
        path
            The path to an attribute
        value
            The value that attribute should have
        """
        self.path = path
        self.value = value



import inspect
import logging
from abc import ABC
import functools
import numpy as np
import time
from typing import Optional, Dict

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.samples.summary import SamplesSummary
from autofit.non_linear.samples.pdf import SamplesPDF
from autofit.non_linear.result import Result
from autofit.non_linear.samples.samples import Samples
from autofit.non_linear.samples.sample import Sample

from .visualize import Visualizer
from ..samples.util import simple_model_for_kwargs

logger = logging.getLogger(__name__)


class Analysis(ABC):
    """
    Protocol for an analysis. Defines methods that can or
    must be implemented to define a class that compute the
    likelihood that some instance fits some data.
    """

    Result = Result
    Visualizer = Visualizer

    LATENT_KEYS = []

    def __init__(
        self, use_jax : bool = False, **kwargs
    ):

        self._use_jax = use_jax
        self.kwargs = kwargs

    def __getattr__(self, item: str):
        """
        If a method starts with 'visualize_' then we assume it is associated with
        the Visualizer and forward the call to the visualizer.

        It may be desirable to remove this behaviour as the visualizer component of
        the system becomes more sophisticated.
        """
        if item.startswith("visualize") or item.startswith("should_visualize"):
            _method = getattr(self.Visualizer, item)
        else:
            raise AttributeError(f"Analysis has no attribute {item}")

        def method(*args, **kwargs):
            parameters = inspect.signature(_method).parameters
            if "analyses" in parameters:
                logger.debug(f"Skipping {item} as this is not a combined analysis")
                return
            return _method(self, *args, **kwargs)

        return method

    @property
    def _xp(self):
        if self._use_jax:
            import jax.numpy as jnp
            return jnp
        return np

    def compute_latent_samples(self, samples: Samples, batch_size : Optional[int] = None) -> Optional[Samples]:
        """
        Compute latent variables from a model instance.

        A latent variable is not itself a free parameter of the model but can be derived from it.
        Latent variables may provide physically meaningful quantities that help interpret a model
        fit, and their values (with errors) are stored in `latent.csv` in parallel with `samples.csv`.

        This implementation is designed to be compatible with both NumPy and JAX:

        - It is written to be side-effect free, so it can be JIT-compiled with `jax.jit`.
        - It can be vectorized over many parameter sets at once using `jax.vmap`, enabling efficient
          batched evaluation of latent variables for multiple samples.
        - Returned values should be simple JAX/NumPy scalars or arrays (no Python objects), so they
          can be stacked into arrays of shape `(n_samples, n_latents)` for batching.
        - Any NaNs introduced (e.g. from invalid model states) can be masked or replaced downstream.

        Parameters
        ----------
        parameters : array-like
            The parameter vector of the model sample. This will typically come from the non-linear search.
            Inside this method it is mapped back to a model instance via `model.instance_from_vector`.
        model : Model
            The model object defining how the parameter vector is mapped to an instance. Passed explicitly
            so that this function can be used inside JAX transforms (`vmap`, `jit`) with `functools.partial`.

        Returns
        -------
        tuple of (float or jax.numpy scalar)
            A tuple containing the latent variables in a fixed order:
            `(intensity_total, magnitude, angle)`. Each entry may be NaN if the corresponding component
            of the model is not present.
        """
        batch_size = batch_size or 10

        try:

            start_latent = time.time()

            compute_latent_for_model = functools.partial(self.compute_latent_variables, model=samples.model)

            if self._use_jax:
                import jax
                start = time.time()
                logger.info("JAX: Applying vmap and jit to likelihood function for latent variables -- may take a few seconds.")
                batched_compute_latent = jax.jit(jax.vmap(compute_latent_for_model))
                logger.info(f"JAX: vmap and jit applied in {time.time() - start} seconds.")
            else:
                def batched_compute_latent(x):
                    return np.array([compute_latent_for_model(xx) for xx in x])

            parameter_array = np.array(samples.parameter_lists)
            latent_samples = []

            # process in batches
            for i in range(0, len(parameter_array), batch_size):

                batch = parameter_array[i:i + batch_size]

                # batched JAX call on this chunk
                latent_values_batch = batched_compute_latent(batch)

                if self._use_jax:
                    import jax.numpy as jnp
                    latent_values_batch = jnp.stack(latent_values_batch, axis=-1)  # (batch, n_latents)
                    mask = jnp.all(jnp.isfinite(latent_values_batch), axis=0)
                    latent_values_batch = latent_values_batch[:, mask]
                else:
                    mask = np.all(np.isfinite(latent_values_batch), axis=0)
                    latent_values_batch = latent_values_batch[:, mask]

                for sample, values in zip(samples.sample_list[i:i + batch_size], latent_values_batch):

                    kwargs = {k: float(v) for k, v in zip(self.LATENT_KEYS, values)}

                    latent_samples.append(
                        Sample(
                            log_likelihood=sample.log_likelihood,
                            log_prior=sample.log_prior,
                            weight=sample.weight,
                            kwargs=kwargs,
                        )
                    )

            print(f"Time to compute latent variables: {time.time() - start_latent} seconds for {len(samples)} samples.")

            return type(samples)(
                sample_list=latent_samples,
                model=simple_model_for_kwargs(latent_samples[0].kwargs),
                samples_info=samples.samples_info,
            )

        except NotImplementedError:
            return None

    def compute_latent_variables(self, parameters, model) -> Dict[str, float]:
        """
        Override to compute latent variables from the instance.

        Latent variables are expressed as a dictionary:
        {"name": value}

        More complex models can be expressed by separating variables
        names by '.'
        {"name.attribute": value}

        Parameters
        ----------
        instance
            An instance of the model.

        Returns
        -------
        The computed latent variables.
        """
        raise NotImplementedError()

    def with_model(self, model):
        """
        Associate an explicit model with this analysis. Instances of the model
        will be used to compute log likelihood in place of the model passed
        from the search.

        Parameters
        ----------
        model
            A model to associate with this analysis

        Returns
        -------
        An analysis for that model
        """
        from .model_analysis import ModelAnalysis

        return ModelAnalysis(analysis=self, model=model)

    def log_likelihood_function(self, instance):
        raise NotImplementedError()

    def save_attributes(self, paths: AbstractPaths):
        pass

    def save_results(self, paths: AbstractPaths, result: Result):
        pass

    def save_results_combined(self, paths: AbstractPaths, result: Result):
        pass

    def modify_before_fit(self, paths: AbstractPaths, model: AbstractPriorModel):
        """
        Overwrite this method to modify the attributes of the `Analysis` class before the non-linear search begins.

        An example use-case is using properties of the model to alter the `Analysis` class in ways that can speed up
        the fitting performed in the `log_likelihood_function`.
        """
        return self

    def modify_model(self, model):
        return model

    def modify_after_fit(
        self, paths: AbstractPaths, model: AbstractPriorModel, result: Result
    ):
        """
        Overwrite this method to modify the attributes of the `Analysis` class before the non-linear search begins.

        An example use-case is using properties of the model to alter the `Analysis` class in ways that can speed up
        the fitting performed in the `log_likelihood_function`.
        """
        return self

    def make_result(
        self,
        samples_summary: SamplesSummary,
        paths: AbstractPaths,
        samples: Optional[SamplesPDF] = None,
        search_internal: Optional[object] = None,
        analysis: Optional[object] = None,
    ) -> Result:
        """
        Returns the `Result` of the non-linear search after it is completed.

        The result type is defined as a class variable in the `Analysis` class. It can be manually overwritten
        by a user to return a user-defined result object, which can be extended with additional methods and attributes
        specific to the model-fit.

        The standard `Result` object may include:

        - The samples summary, which contains the maximum log likelihood instance and median PDF model.

        - The paths of the search, which are used for loading the samples and search internal below when a search
        is resumed.

        - The samples of the non-linear search (e.g. MCMC chains) also stored in `samples.csv`.

        - The non-linear search used for the fit in its internal representation, which is used for resuming a search
        and making bespoke visualization using the search's internal results.

        - The analysis used to fit the model (default disabled to save memory, but option may be useful for certain
        projects).

        Parameters
        ----------
        samples_summary
            The summary of the samples of the non-linear search, which include the maximum log likelihood instance and
            median PDF model.
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        samples
            The samples of the non-linear search, for example the chains of an MCMC run.
        search_internal
            The internal representation of the non-linear search used to perform the model-fit.
        analysis
            The analysis used to fit the model.

        Returns
        -------
        Result
            The result of the non-linear search, which is defined as a class variable in the `Analysis` class.
        """
        return self.Result(
            samples_summary=samples_summary,
            paths=paths,
            samples=samples,
            search_internal=search_internal,
            analysis=analysis,
        )

    def profile_log_likelihood_function(self, paths: AbstractPaths, instance):
        """
        Overwrite this function for profiling of the log likelihood function to be performed every update of a
        non-linear search.

        This behaves analogously to overwriting the `visualize` function of the `Analysis` class, whereby the user
        fills in the project-specific behaviour of the profiling.

        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        instance
            The maximum likliehood instance of the model so far in the non-linear search.
        """
        pass

    def perform_quick_update(self, paths, instance):
        raise NotImplementedError

    def print_vram_use(self, model, batch_size : int) -> str:
        """
        Print JAX VRAM use for a given batch size.

        Parameters
        ----------
        batch_size
            The batch size to profile, which is the number of model evaluations JAX will perform simultaneously.
        """
        if not self._use_jax:
            print("use_jax=False for this analysis, therefore does not use GPU and VRAM use cannot be profiled.")
            return

        import jax
        import jax.numpy as jnp

        from autofit.non_linear.fitness import Fitness

        fitness = Fitness(
            model=model,
            analysis=self,
            fom_is_log_likelihood=True,
            use_jax_vmap=True,
            batch_size=batch_size,
        )

        parameters = np.zeros((batch_size, model.total_free_parameters))

        for i in range(batch_size):
            parameters[i, :] = model.physical_values_from_prior_medians

        parameters = jnp.array(parameters)

        batched_call = jax.jit(jax.vmap(fitness.call))
        lowered = batched_call.lower(parameters)
        compiled = lowered.compile()
        memory_analysis = compiled.memory_analysis()

        vram_bytes = (
                memory_analysis.output_size_in_bytes
                + memory_analysis.temp_size_in_bytes
        )

        if vram_bytes == 0:
            print(
                "VRAM USE = 0.000 GB "
                "(this likely means JAX is running in CPU-only mode)"
            )
        else:
            print(
                f"VRAM USE = {vram_bytes / 1024 ** 3:.3f} GB"
            )


from typing import List, cast, Type

import numpy as np

from autoconf.class_path import get_class
from .sample import Sample
from .samples import Samples
from .pdf import SamplesPDF


class EfficientSamples:
    def __init__(self, samples: SamplesPDF):
        """
        A representation of samples where values are stored in numpy arrays. This is more
        efficient for storage and computation.

        Parameters
        ----------
        samples
            A representation of samples where values are in instances of the Sample class.
        """
        self.model = samples.model
        self.samples_info = samples.samples_info

        sample_list = samples.sample_list
        try:
            self._keys = list(sample_list[0].kwargs.keys())
        except IndexError:
            self._keys = []
        self._values = np.asarray(
            [[sample.kwargs[key] for key in self._keys] for sample in sample_list]
        )
        self._log_likelihoods = np.asarray(
            [sample.log_likelihood for sample in sample_list]
        )
        self._log_priors = np.asarray([sample.log_prior for sample in sample_list])
        self._weights = np.asarray([sample.weight for sample in sample_list])

    @property
    def samples(self) -> Samples:
        """
        Convert the efficient samples back to a SamplesPDF instance.
        """
        cls = cast(Type[Samples], get_class(self.samples_info["class_path"]))
        return cls(
            model=self.model,
            samples_info=self.samples_info,
            sample_list=self.sample_list,
        )

    @property
    def sample_list(self) -> List[Sample]:
        """
        Convert the efficient samples back to a list of Sample instances.
        """
        return [
            Sample(
                log_likelihood=log_likelihood,
                log_prior=log_prior,
                weight=weight,
                kwargs=kwargs,
            )
            for log_likelihood, log_prior, weight, kwargs in zip(
                self._log_likelihoods,
                self._log_priors,
                self._weights,
                self._kwargs,
            )
        ]

    @property
    def _kwargs(self) -> List[dict]:
        """
        Convert the efficient samples back to a list of dictionaries of kwargs.
        """
        return [
            {key: value for key, value in zip(self._keys, values)}
            for values in self._values
        ]



import copy
import logging
import os
from os import path
from typing import List, Tuple, Union, Type, Optional, Dict

from autoconf.dictable import to_dict
from autofit import exc
from autofit.mapper import prior as p
from autofit.non_linear.parallel import Process
from autofit.text.formatter import write_table
from .job import Job
from .result import GridSearchResult
from .result_builder import ResultBuilder


class Sequential:
    # noinspection PyUnusedLocal
    @staticmethod
    def run_jobs(jobs, *_, **kwargs):
        for job_ in jobs:
            yield job_.perform()


class GridSearch:
    def __init__(
        self,
        search,
        number_of_steps: int = 4,
        number_of_cores: int = 1,
        result_output_interval: int = 100,
    ):
        """
        Performs a non linear optimiser search for each square in a grid. The dimensionality of the search depends on
        the number of distinct priors passed to the fit function. (1 / step_size) ^ no_dimension steps are performed
        per an optimisation.

        Parameters
        ----------
        number_of_steps
            The number of steps to go in each direction
        search
            The class of the search that is run at each step
        result_output_interval
            The interval between saving a GridSearchResult object via paths
        """
        self.number_of_steps = number_of_steps
        self.search = search

        self._logger = None

        self.logger.info("Creating grid search")

        self.number_of_cores = number_of_cores or 1

        self.logger.info(f"Using {number_of_cores} core(s)")

        self._result_output_interval = result_output_interval

    __exclude_identifier_fields__ = ("number_of_cores",)

    def __getstate__(self):
        """
        Remove the logger for pickling
        """
        state = self.__dict__.copy()
        if "_logger" in state:
            del state["_logger"]
        return state

    @property
    def logger(self):
        if not hasattr(self, "_logger") or self._logger is None:
            self._logger = logging.getLogger(f"GridSearch ({self.search.name})")
        return self._logger

    @property
    def paths(self):
        return self.search.paths

    @property
    def parallel(self) -> bool:
        """
        Should this grid search run in parallel?
        """
        return self.number_of_cores > 1

    @property
    def step_size(self):
        """
        Returns
        -------
        step_size: float
            The size of a step in any given dimension in hyper space.
        """
        return 1 / self.number_of_steps

    def make_physical_lists(self, grid_priors) -> List[List[float]]:
        lists = self.make_lists(grid_priors)
        return [
            [prior.value_for(value) for prior, value in zip(grid_priors, l)]
            for l in lists
        ]

    def make_lists(self, grid_priors):
        """
        Produces a list of lists of floats, where each list of floats represents the values in each dimension for one
        step of the grid search.

        Parameters
        ----------
        grid_priors: [p.Prior]
            A list of priors that are to be searched using the grid search.

        Returns
        -------
        lists: [[float]]
        """
        return make_lists(
            len(grid_priors), step_size=self.step_size, centre_steps=False
        )

    def make_arguments(self, values, grid_priors):
        arguments = {}
        for value, grid_prior in zip(values, grid_priors):
            try:
                grid_prior.lower_limit
                grid_prior.upper_limit
            except AttributeError:
                raise exc.PriorException(
                    "Priors passed to the grid search must have upper and lower limit (e.g. be UniformPrior)"
                )
            if (
                float("-inf") == grid_prior.lower_limit
                or float("inf") == grid_prior.upper_limit
            ):
                raise exc.PriorException(
                    "Priors passed to the grid search must have definite limits"
                )
            lower_limit = grid_prior.lower_limit + value * grid_prior.width
            upper_limit = (
                grid_prior.lower_limit + (value + self.step_size) * grid_prior.width
            )
            prior = p.UniformPrior(lower_limit=lower_limit, upper_limit=upper_limit)
            arguments[grid_prior] = prior
        return arguments

    def model_mappers(self, model, grid_priors):
        grid_priors = model.sort_priors_alphabetically(set(grid_priors))
        lists = self.make_lists(grid_priors)
        for values in lists:
            arguments = self.make_arguments(values, grid_priors)
            yield model.mapper_from_partial_prior_arguments(arguments)

    def fit(
        self,
        model,
        analysis,
        grid_priors,
        info: Optional[Dict] = None,
    ):
        """
        Fit an analysis with a set of grid priors. The grid priors are priors associated with the model mapper
        of this instance that are replaced by uniform priors for each step of the grid search.

        Parameters
        ----------
        info
        model
        analysis: autofit.non_linear.non_linear.Analysis
            An analysis used to determine the fitness of a given model instance
        grid_priors: [p.Prior]
            A list of priors to be substituted for uniform priors across the grid.

        Returns
        -------
        result: GridSearchResult
            An object that comprises the results from each individual fit
        """
        self.paths.model = model
        self.paths.search = self

        self.logger.info("Running grid search...")

        process_class = Process if self.parallel else Sequential
        # noinspection PyArgumentList
        return self._fit(
            model=model,
            analysis=analysis,
            grid_priors=grid_priors,
            process_class=process_class,
            info=info,
        )

    def _fit(
        self,
        model,
        analysis,
        grid_priors,
        process_class=Union[Type[Process], Type[Sequential]],
        info: Optional[Dict] = None,
    ):
        """
        Perform the grid search in parallel, with all the optimisation for each grid square being performed on a
        different process.

        Parameters
        ----------
        model
        analysis
            An analysis
        grid_priors
            Priors describing the position in the grid
        process_class
            Class used to dispatch jobs

        Returns
        -------
        result: GridSearchResult
            The result of the grid search
        """
        grid_priors = model.sort_priors_alphabetically(set(grid_priors))
        lists = self.make_lists(grid_priors)

        jobs = self.make_jobs(model, analysis, grid_priors, info)

        builder = ResultBuilder(
            lists=lists,
            grid_priors=grid_priors,
            paths=[j.search_instance.paths for j in jobs],
        )

        self.save_metadata()

        def save_results():
            self.paths.save_json("result", to_dict(builder()))

        def write_results():
            self.logger.debug("Writing results")

            os.makedirs(self.paths.output_path, exist_ok=True)

            is_evidence = any(row[-1] is not None for row in results_list)

            headers = [
                "index",
                *map(model.name_for_prior, grid_priors),
                "log_likelihood_increase",
            ]
            rows = results_list

            if is_evidence:
                headers.append("log_evidence")
            else:
                rows = [row[:-1] for row in rows]

            write_table(
                headers,
                rows,
                self.paths.output_path / "results.csv",
            )

        results_list = []

        for i, job_result in enumerate(
            process_class.run_jobs(jobs, self.number_of_cores)
        ):
            builder.add(job_result)
            results_list.append(job_result.result_list_row)
            write_results()
            if i % self._result_output_interval == 0:
                save_results()

        save_results()
        self.paths.completed()

        return builder()

    def save_metadata(self):
        self.paths.save_unique_tag(is_grid_search=True)

    def make_jobs(self, model, analysis, grid_priors, info: Optional[Dict] = None):
        grid_priors = model.sort_priors_alphabetically(set(grid_priors))
        lists = self.make_lists(grid_priors)

        jobs = list()

        for index, values in enumerate(lists):
            jobs.append(
                self.job_for_analysis_grid_priors_and_values(
                    analysis=copy.deepcopy(analysis),
                    model=model,
                    grid_priors=grid_priors,
                    values=values,
                    index=index,
                    info=info,
                )
            )
        return jobs

    def job_for_analysis_grid_priors_and_values(
        self, model, analysis, grid_priors, values, index, info: Optional[Dict] = None
    ):
        arguments = self.make_arguments(values=values, grid_priors=grid_priors)
        model = model.mapper_from_partial_prior_arguments(arguments=arguments)

        labels = []
        for prior in sorted(arguments.values(), key=lambda pr: pr.id):
            labels.append(
                "{}_{:.2f}_{:.2f}".format(
                    model.name_for_prior(prior), prior.lower_limit, prior.upper_limit
                )
            )

        name_path = path.join(
            self.paths.name,
            self.paths.identifier,
            "_".join(labels),
        )

        search_instance = self.search_instance(name_path=name_path)
        search_instance.paths.model = model

        return Job(
            search_instance=search_instance,
            model=model,
            analysis=analysis,
            arguments=arguments,
            index=index,
            info=info,
        )

    def search_instance(self, name_path):
        search_instance = self.search.copy_with_paths(
            self.paths.create_child(
                name=name_path,
                path_prefix=self.paths.path_prefix,
                is_identifier_in_paths=False,
            )
        )

        for key, value in self.__dict__.items():
            if key not in ("model", "instance", "paths", "search", "number_of_cores"):
                try:
                    setattr(search_instance, key, value)
                except AttributeError:
                    pass

        if self.number_of_cores > 1:
            search_instance.number_of_cores = 1

        return search_instance


def grid(fitness, no_dimensions, step_size):
    """
    Grid2D search using a fitness function over a given number of dimensions and a given step size between inclusive
    limits of 0 and 1.

    Parameters
    ----------
    fitness: function
        A function that takes a tuple of floats as an argument
    no_dimensions: int
        The number of dimensions of the grid search
    step_size: float
        The step size of the grid search

    Returns
    -------
    best_arguments: tuple[float]
        The tuple of arguments that gave the highest fitness
    """
    best_fitness = float("-inf")
    best_arguments = None

    for arguments in make_lists(no_dimensions, step_size):
        fitness = fitness(tuple(arguments))
        if fitness > best_fitness:
            best_fitness = fitness
            best_arguments = tuple(arguments)

    return best_arguments


def make_lists(
    no_dimensions: int, step_size: Union[Tuple[float], float], centre_steps=True
):
    """
        Returns a list of lists of floats covering every combination across no_dimensions of points of integer step size
    between 0 and 1 inclusive.

    Parameters
    ----------
    no_dimensions
        The number of dimensions, that is the length of the lists
    step_size
        The step size. This can be a float or a tuple with the same number of dimensions
    centre_steps

    Returns
    -------
    lists: [[float]]
        A list of lists
    """
    if isinstance(step_size, float):
        step_size = tuple(step_size for _ in range(no_dimensions))

    if no_dimensions == 0:
        return [[]]

    sub_lists = make_lists(no_dimensions - 1, step_size[1:], centre_steps=centre_steps)
    step_size = step_size[0]
    return [
        [step_size * value + (0.5 * step_size if centre_steps else 0)] + sub_list
        for value in range(int((1 / step_size)))
        for sub_list in sub_lists
    ]



from typing import Optional

from autoconf import conf
from autofit.database.sqlalchemy_ import sa
from autofit.non_linear.search.abstract_search import NonLinearSearch
from autofit.non_linear.initializer import Initializer
from autofit.non_linear.samples import SamplesMCMC
from autofit.non_linear.search.mcmc.auto_correlations import AutoCorrelationsSettings
from autofit.non_linear.plot.mcmc_plotters import MCMCPlotter
from autofit.non_linear.plot.output import Output

class AbstractMCMC(NonLinearSearch):

    def __init__(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            unique_tag: Optional[str] = None,
            initializer: Optional[Initializer] = None,
            auto_correlation_settings=AutoCorrelationsSettings(),
            iterations_per_full_update: Optional[int] = None,
            iterations_per_quick_update: int = None,
            number_of_cores: Optional[int] = None,
            session: Optional[sa.orm.Session] = None,
            **kwargs
    ):
        
        self.auto_correlation_settings = auto_correlation_settings
        self.auto_correlation_settings.update_via_config(
            config=self.config_type[self.__class__.__name__]["auto_correlations"]
        )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            initializer=initializer,
            iterations_per_quick_update=iterations_per_quick_update,
            iterations_per_full_update=iterations_per_full_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

    @property
    def config_type(self):
        return conf.instance["non_linear"]["mcmc"]

    @property
    def samples_cls(self):
        return SamplesMCMC

    @property
    def plotter_cls(self):
        return MCMCPlotter

    def plot_results(self, samples):

        if not samples.pdf_converged:
            return

        def should_plot(name):
            return conf.instance["visualize"]["plots_search"]["mcmc"][name]

        plotter = self.plotter_cls(
            samples=samples,
            output=Output(path=self.paths.image_path / "search", format="png"),
        )

        if should_plot("corner_cornerpy"):
            plotter.corner_cornerpy()



from typing import List, Tuple, Union, Dict

from autofit.non_linear.grid.grid_list import GridList, as_grid_list
from autofit.non_linear.grid.grid_search.result import AbstractGridSearchResult
from autofit.non_linear.samples.interface import SamplesInterface


# noinspection PyTypeChecker
class SensitivityResult(AbstractGridSearchResult):
    def __init__(
        self,
        samples: List[SamplesInterface],
        perturb_samples: List[SamplesInterface],
        shape: Tuple[int, ...],
        path_values: Dict[Tuple[str, ...], List[float]],
    ):
        """
        The result of a sensitivity mapping

        Parameters
        ----------
        shape
            The shape of the sensitivity mapping grid.
        path_values
            A list of tuples of the path to the grid priors and the physical values themselves.
        """
        super().__init__(GridList(samples, shape))
        self.perturb_samples = GridList(perturb_samples, shape)
        self.shape = shape
        self.path_values = path_values

    def perturbed_physical_centres_list_from(
        self, path: Union[str, Tuple[str, ...]]
    ) -> GridList:
        """
        Returns the physical centres of the perturbed model for each sensitivity fit

        Parameters
        ----------
        path
            The path to the physical centres in the samples
        """
        if isinstance(path, str):
            path = tuple(path.split("."))
        return self.path_values[path]

    def __getitem__(self, item):
        return self.samples[item]

    def __iter__(self):
        return iter(self.samples)

    def __len__(self):
        return len(self.samples)

    @property
    @as_grid_list
    def log_evidences_base(self) -> GridList:
        """
        The log evidences of the base model for each sensitivity fit
        """
        return [sample.log_evidence for sample in self.samples]

    @property
    @as_grid_list
    def log_evidences_perturbed(self) -> GridList:
        """
        The log evidences of the perturbed model for each sensitivity fit
        """
        return [sample.log_evidence for sample in self.perturb_samples]

    @property
    @as_grid_list
    def log_evidence_differences(self) -> GridList:
        """
        The log evidence differences between the base and perturbed models
        """
        return [
            log_evidence_perturbed - log_evidence_base
            for log_evidence_perturbed, log_evidence_base in zip(
                self.log_evidences_perturbed, self.log_evidences_base
            )
        ]

    @property
    @as_grid_list
    def log_likelihoods_base(self) -> GridList:
        """
        The log likelihoods of the base model for each sensitivity fit
        """
        return [sample.log_likelihood for sample in self.samples]

    @property
    @as_grid_list
    def log_likelihoods_perturbed(self) -> GridList:
        """
        The log likelihoods of the perturbed model for each sensitivity fit
        """
        return [sample.log_likelihood for sample in self.perturb_samples]

    @property
    @as_grid_list
    def log_likelihood_differences(self) -> GridList:
        """
        The log likelihood differences between the base and perturbed models
        """
        return [
            log_likelihood_perturbed - log_likelihood_base
            for log_likelihood_perturbed, log_likelihood_base in zip(
                self.log_likelihoods_perturbed, self.log_likelihoods_base
            )
        ]

    def figure_of_merits(
        self,
        use_log_evidences: bool,
    ) -> GridList:
        """
        Convenience method to get either the log likelihoods difference or log evidence difference of the grid search.

        Parameters
        ----------
        use_log_evidences
            If true, the log evidences are returned, otherwise the log likelihoods are returned.
        """

        if use_log_evidences:
            return self.log_evidence_differences
        return self.log_likelihood_differences



import copy
from abc import ABC, abstractmethod
from typing import List, Dict, cast

from autofit.mapper.model import ModelInstance

from .query import InterpolatorPath, Equality
from .. import Model


class AbstractInterpolator(ABC):
    def __init__(self, instances: List[ModelInstance]):
        """
        A TimeSeries allows interpolation on any variable.

        For example, each instance may have an attribute t. Other attributes
        can be determined for any given value of t by interpolating their
        values for each instance in the time series.

        Parameters
        ----------
        instances
            A list of instances (e.g. best fits) each with all the same attributes
            and a value on which a time series may be built
        """
        self.instances = instances

    def __getattr__(self, item: str) -> InterpolatorPath:
        """
        Used to indicate which attribute is the time attribute.

        This attribute may be at any path in the model.

        e.g.
        instance.t
        instance.something.t

        Parameters
        ----------
        item
            The name of an attribute of the instance

        Returns
        -------
        A class that keeps track of which attributes have been addressed
        """
        return InterpolatorPath([item])

    def _value_map(self, path: InterpolatorPath) -> Dict[float, ModelInstance]:
        """
        Maps know values to corresponding instances for a given path

        Parameters
        ----------
        path
            A path to an attribute, e.g. time

        Returns
        -------
        A dictionary mapping values of that attribute for each instance to the corresponding
        instance
        """
        return {path.get_value(instance): instance for instance in self.instances}

    def __getitem__(self, item: Equality) -> ModelInstance:
        """
        Create an artificial model instance which has values interpolated
        for a given interpolation value.

        Parameters
        ----------
        item
            Indicates a value for a given attribute to which the instance should
            be interpolated

        Returns
        -------
        An artificial instance with values interpolated

        Examples
        --------
        # Each instance in the time_series has an attribute 't'
        time_series = af.LinearTimeSeries([instance_1, instance_2, instance_3])

        # We can now create an instance which has a value of t = 3.5 by interpolating
        instance = time_series[time_series.t == 3.5)

        # We can also interpolate on any arbitrary attribute of the instance
        instance = time_series[time_series.some.arbitrary.attribute == -1.0]
        """
        value_map = self._value_map(item.path)
        try:
            return value_map[item.value]
        except KeyError:
            value_map = self._value_map(item.path)
            x = sorted(value_map)

            instance = self.instances[0]
            new_instance = copy.copy(instance)

            for path, _ in instance.path_instance_tuples_for_class(float):
                y = [value_map[value].object_for_path(path) for value in x]

                new_instance = new_instance.replacing_for_path(
                    path,
                    self._interpolate(x, cast(List[float], y), item.value),
                )

        # noinspection PyTypeChecker
        return new_instance.replacing_for_path(tuple(item.path.keys), item.value)

    def relationships(self, path: InterpolatorPath) -> Model:
        instance = self.instances[0]
        new_instance = copy.copy(instance)

        value_map = self._value_map(path)
        x = sorted(value_map)

        for path, _ in instance.path_instance_tuples_for_class(float):
            y = [value_map[value].object_for_path(path) for value in x]

            new_instance = new_instance.replacing_for_path(
                path,
                self._relationship(x, cast(List[float], y)),
            )

        return new_instance

    @classmethod
    def _interpolate(cls, x: List[float], y: List[float], value: float) -> float:
        """
        Interpolate a given attribute to find its effective value at some time

        Parameters
        ----------
        x
            A list of times (or another series)
        y
            A list which one value for each time
        value
            The time for which we want an interpolated value for the attribute

        Returns
        -------
        An interpolated value for the attribute
        """
        return cls._relationship(x, y)(value)

    @staticmethod
    @abstractmethod
    def _relationship(x: List[float], y: List[float]):
        """
        Calculate the relationship between two values

        Parameters
        ----------
        x
            The first value
        y
            The second value

        Returns
        -------
        The relationship between the two values
        """



import re
from enum import Enum
from typing import Dict, List, Union

from pathlib import Path

from autofit.aggregator.search_output import SearchOutput
from autofit.aggregator import Aggregator


def subplot_filename(subplot: Enum) -> str:
    subplot_type = subplot.__class__
    return (
        re.sub(
            r"([A-Z])",
            r"_\1",
            subplot_type.__name__.replace("FITS", ""),
        )
        .lower()
        .lstrip("_")
    )

class AggregateFITS:
    def __init__(self, aggregator: Union[Aggregator, List[SearchOutput]]):
        """
        A class for extracting fits files from the aggregator.

        Parameters
        ----------
        aggregator
            The aggregator containing the fits files.
        """
        if len(aggregator) == 0:
            raise ValueError("The aggregator is empty.")

        self.aggregator = aggregator

    @staticmethod
    def _hdus(
        result: SearchOutput,
        hdus: List[Enum],
        extname_prefix = None
    ) -> "List[fits.ImageHDU]":
        """
        Extract the HDUs from a given fits for a given search.

        Parameters
        ----------
        result
            The search output.
        hdus
            The HDUs to extract.

        Returns
        -------
        The extracted HDUs.
        """
        from astropy.io import fits

        row = []
        for hdu in hdus:
            source = result.value(subplot_filename(hdu))
            source_hdu = source[source.index_of(hdu.value)]

            if extname_prefix is not None:
                source_hdu.header["EXTNAME"] = f"{extname_prefix.upper()}_{source_hdu.header['EXTNAME']}"

            row.append(
                fits.ImageHDU(
                    data=source_hdu.data,
                    header=source_hdu.header,
                )
            )
        return row

    def extract_fits(self, hdus: List[Enum], extname_prefix_list = None) -> "fits.HDUList":
        """
        Extract the HDUs from the fits files for every search in the aggregator.

        Return the result as a list of HDULists. The first HDU in each list is an empty PrimaryHDU.

        Parameters
        ----------
        hdus
            The HDUs to extract.

        Returns
        -------
        The extracted HDUs.
        """
        from astropy.io import fits

        output = [fits.PrimaryHDU()]
        for i, result in enumerate(self.aggregator):

            extname_prefix = extname_prefix_list[i] if extname_prefix_list is not None else None

            output.extend(self._hdus(result, hdus, extname_prefix))

        return fits.HDUList(output)

    def extract_csv(self, filename: str) -> List[Dict]:
        """
        Extract .csv files which store imaging results that are typically on irregular grids and thus don't suit
        a .fits file.

        Return the result as a list of Dicts corresponding to the fits in the aggregator.

        Parameters
        ----------
        filename
            The name of the .csv file to extract.

        Returns
        -------
        The extracted HDUs.
        """
        from astropy.table import Table

        output = []
        for result in self.aggregator:
            output.append(Table.read(result.value(filename), format="fits"))

        return output

    def output_to_folder(
        self,
        folder: Path,
        name: Union[str, List[str]],
        hdus: List[Enum],
    ):
        """
        Output the fits files for every search in the aggregator to a folder.

        Only include HDUs specific in the hdus argument.

        Parameters
        ----------
        folder
            The folder to output the fits files to.
        hdus
            The HDUs to output.
        name
            The name of the fits file. This is the attribute of the search output that is used to name the file.
            OR a list of names for each HDU.
        """
        from astropy.io import fits

        folder.mkdir(parents=True, exist_ok=True)

        for i, result in enumerate(self.aggregator):
            if isinstance(name, str):
                output_name = getattr(result, name)
            else:
                output_name = name[i]

            hdu_list = fits.HDUList(
                [fits.PrimaryHDU()]
                + self._hdus(
                    result,
                    hdus,
                )
            )
            with open(folder / f"{output_name}.fits", "wb") as file:
                hdu_list.writeto(file)



from functools import wraps
from typing import List, Tuple

import numpy as np


def as_grid_list(func):
    """
    Wrap functions with a function which converts the output list of grid search results to a `GridList` object.

    Parameters
    ----------
    func
        A function which computes and retrusn a list of grid search results.

    Returns
    -------
        A function which converts a list of grid search results to a `GridList` object.
    """

    @wraps(func)
    def wrapper(grid_search_result, *args, **kwargs) -> List:
        """
        This decorator converts the output of a function which computes a list of grid search results to a `GridList`.

        Parameters
        ----------
        grid_search_result
            The instance of the `GridSearchResult` which is being operated on.

        Returns
        -------
            The function output converted to a `GridList`.
        """

        values = func(grid_search_result, *args, **kwargs)

        return GridList(values=values, shape=grid_search_result.shape)

    return wrapper


class GridList(list):
    def __init__(self, values: List, shape: Tuple):
        """
        Many quantities of a `GridSearchResult` are stored as lists of lists.

        The number of lists corresponds to the dimensionality of the grid search and the number of elements
        in each list corresponds to the number of steps in that grid search dimension.

        This class provides a wrapper around lists of lists to provide some convenience methods for accessing
        the values in the lists. For example, it provides a conversion of the list of list structure to a ndarray.

        For example, for a 2x2 grid search the shape of the Numpy array is (2,2) and it is numerically ordered such
        that the first search's entries(corresponding to unit priors (0.0, 0.0)) are in the first
        value (E.g. entry [0, 0]) of the NumPy array.

        Parameters
        ----------
        values
        """
        super().__init__(values)

        self.shape = shape

    @property
    def as_list(self) -> List:
        return self

    @property
    def native(self) -> np.ndarray:
        """
        The list of lists as an ndarray.
        """
        return np.reshape(np.array(self), self.shape)



import numpy as np

import autofit as af

class ResultExample(af.Result):

    @property
    def max_log_likelihood_model_data_1d(self) -> np.ndarray:
        """
        Returns the maximum log likelihood model's 1D model data.

        This is an example of how we can pass the `Analysis` class a custom `Result` object and extend this result
        object with new properties that are specific to the model-fit we are performing.
        """
        return self.analysis.model_data_1d_from(instance=self.instance)



from .gaussian import GaussianPrior
from .truncated_gaussian import TruncatedGaussianPrior
from .log_gaussian import LogGaussianPrior
from .log_uniform import LogUniformPrior
from .uniform import UniformPrior



from autofit.database.migration import Step, Migrator

steps = [
    Step(
        "ALTER TABLE fit ADD name VARCHAR;", "ALTER TABLE fit ADD path_prefix VARCHAR;"
    ),
    Step(
        "CREATE TABLE named_instance (id INTEGER NOT NULL , name VARCHAR, fit_id VARCHAR, PRIMARY KEY (id), FOREIGN KEY (fit_id) REFERENCES fit (id));"
    ),
    Step("ALTER TABLE fit ADD max_log_likelihood FLOAT;"),
    Step(
        "CREATE TABLE json (id INTEGER NOT NULL, name VARCHAR, string VARCHAR, fit_id VARCHAR, PRIMARY KEY (id), FOREIGN KEY (fit_id) REFERENCES fit (id));"
    ),
    Step(
        "CREATE TABLE array (id INTEGER NOT NULL, name VARCHAR, bytes BLOB, _dtype VARCHAR, _shape VARCHAR, array_type VARCHAR, fit_id VARCHAR, PRIMARY KEY (id), FOREIGN KEY (fit_id) REFERENCES fit (id));"
    ),
    Step(
        "CREATE TABLE hdu (id INTEGER NOT NULL, array_type VARCHAR, _header VARCHAR, PRIMARY KEY (id), FOREIGN KEY (id) REFERENCES array (id));"
    ),
    Step(
        "ALTER TABLE object ADD COLUMN latent_variables_for_id INTEGER;",
    ),
    Step(
        "ALTER TABLE object RENAME COLUMN latent_variables_for_id TO latent_samples_for_id;",
    ),
    Step(
        "CREATE TABLE fits (id INTEGER NOT NULL, name VARCHAR, fit_id VARCHAR, PRIMARY KEY (id), FOREIGN KEY (fit_id) REFERENCES fit (id));"
    ),
    Step(
        "ALTER TABLE hdu ADD COLUMN fits_id INTEGER;",
    ),
    Step(
        "ALTER TABLE hdu ADD COLUMN is_primary BOOLEAN;",
    ),
]

migrator = Migrator(*steps)



from abc import ABC, abstractmethod
from functools import wraps
from typing import Union, List, Tuple, Dict

from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior_model.abstract import Path


def to_instance(func):
    """
    Decorator for methods that return a vector of parameters, which can be converted to a model instance.

    Parameters
    ----------
    func
        A method that returns a vector of parameters

    Returns
    -------
    A wrapper that converts the vector to a model instance
    """

    @wraps(func)
    def wrapper(
        self,
        *args,
        as_instance: bool = True,
        as_dict: bool = False,
        **kwargs,
    ) -> Union[List, Dict, ModelInstance]:
        vector = func(self, *args, **kwargs)

        if as_dict:
            return {".".join(path[0]): value for path, value in zip(self.paths, vector)}

        if as_instance:
            return self._instance_from_vector(vector)

        return vector

    return wrapper


class SamplesInterface(ABC):
    def __init__(self, model: AbstractPriorModel):
        """
        An interface for the samples of a `NonLinearSearch` that has been run, including the maximum log likelihood
        """
        self.model = model

        self._paths = None
        self._names = None
        self._instance = None

    @property
    def instance(self):
        if self._instance is None:
            self._instance = self.max_log_likelihood()
        return self._instance

    @property
    def paths(self) -> List[Tuple[Path]]:
        """
        A list of paths to unique priors in the same order as prior
        ids (and therefore sample columns)

        Uses hasattr to make backwards compatible
        """
        if not hasattr(self, "_paths") or self._paths is None:
            self._paths = self.model.all_paths
        return self._paths

    @property
    def names(self) -> List[Tuple[str]]:
        """
        A list of names of unique priors in the same order as prior
        ids (and therefore sample columns)

        Uses hasattr to make backwards compatible
        """
        if not hasattr(self, "_names") or self._names is None:
            self._names = self.model.all_names
        return self._names

    @to_instance
    def max_log_likelihood(self, as_instance: bool = True) -> List[float]:
        """
        The parameters of the maximum log likelihood sample of the `NonLinearSearch` returned as a model instance or
        list of values.
        """

        sample = self.max_log_likelihood_sample

        return sample.parameter_lists_for_paths(
            self.paths if sample.is_path_kwargs else self.names
        )

    @property
    @abstractmethod
    def max_log_likelihood_sample(self):
        pass

    @property
    def log_likelihood(self):
        return self.max_log_likelihood_sample.log_likelihood

    @property
    @abstractmethod
    def log_evidence(self):
        pass

    @property
    def model_centred(self) -> AbstractPriorModel:
        """
        Returns a model where every free parameter is a `GaussianPrior` with `mean` the previous result's
        inferred maximum log likelihood parameter values and `sigma` values determined from the errors on
        each parameter or the prior config files

        For example, a previous result may infer a parameter to have a maximum log likelihood value of 2.

        If this result is used for search chaining, `model_centred` will assign this free parameter
        `GaussianPrior(mean=2.0, sigma=0.1)` in the new model, where `sigma` is determined from the errors on
        each parameter or the prior config files.

        Returns
        -------
        A model mapper created by taking results from this search.
        """
        return self.model.mapper_from_prior_means(means=self.prior_means)

    def model_centred_absolute(self, a: float) -> AbstractPriorModel:
        """
        Returns a model where every free parameter is a `GaussianPrior` with `mean` the previous result's
        inferred maximum log likelihood parameter values and `sigma` the input absolute value `a`.

        For example, a previous result may infer a parameter to have a maximum log likelihood value of 2.

        If this result is used for search chaining, `model_centred_absolute(a=0.1)` will assign this free parameter
        `GaussianPrior(mean=2.0, sigma=0.1)` in the new model, where `sigma` is linked to the input `a`.

        Parameters
        ----------
        a
            The absolute width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this search and creating priors with the defined absolute
        width.
        """
        return self.model.mapper_from_prior_means(means=self.prior_means, a=a)

    def model_centred_relative(self, r: float) -> AbstractPriorModel:
        """
        Returns a model where every free parameter is a `GaussianPrior` with `mean` the previous result's
        inferred maximum log likelihood parameter values and `sigma` a relative value from the result `r`.

        For example, a previous result may infer a parameter to have a maximum log likelihood value of 2 and
        an error at the input `sigma` of 0.5.

        If this result is used for search chaining, `model_centred_relative(r=0.1)` will assign this free parameter
        `GaussianPrior(mean=2.0, sigma=0.5*0.1)` in the new model, where `sigma` is the inferred error times `r`.

        Parameters
        ----------
        r
            The relative width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this search and creating priors with the defined relative
        width.
        """
        return self.model.mapper_from_prior_means(means=self.prior_means, r=r)

    def model_centred_max_lh_bounded(self, b: float) -> AbstractPriorModel:
        """
        Returns a model where every free parameter is a `UniformPrior` with `lower_limit` and `upper_limit the previous
        result's inferred maximum log likelihood parameter values minus and plus the bound `b`.

        For example, a previous result may infer a parameter to have a maximum log likelihood value of 2.

        If this result is used for search chaining, `model_bound(b=0.1)` will assign this free parameter
        `UniformPrior(lower_limit=1.9, upper_limit=2.1)` in the new model.

        Parameters
        ----------
        b
            The size of the bounds of the uniform prior

        Returns
        -------
        A model mapper created by taking results from this search and creating priors with the defined bounded
        uniform priors.
        """
        return self.model.mapper_from_uniform_floats(
            floats=self.max_log_likelihood(as_instance=False), b=b
        )

    def _instance_from_vector(self, vector: List[float]) -> ModelInstance:
        return self.model.instance_from_vector(vector=vector)

    @property
    def prior_means(self) -> [List]:
        """
        The mean of every parameter used to link its inferred values and errors to priors used to sample the
        same (or similar) parameters in a subsequent search, where:

        - The mean is given by their most-probable values median PDF values if using a sampler that provides this
          information(using median_pdf(as_instance=False)).
        - The man is given by the maximum log likelihood values otherwise (e.g. for a maximum likelihood estimator).
        """

        try:
            return self.median_pdf(as_instance=False)
        except (AttributeError, IndexError):
            return self.max_log_likelihood(as_instance=False)

    def path_map_for_model(self, model):
        return {
            tuple(self.model.all_paths_for_prior(prior)): path
            for path, prior in model.path_priors_tuples
        }



from .model import Object
from ..sqlalchemy_ import sa


class Compound(Object):
    __tablename__ = "compound"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )
    compound_type = sa.Column(sa.String)

    @property
    def left(self):
        return self.children[0]

    @property
    def right(self):
        return self.children[1]

    __mapper_args__ = {"polymorphic_identity": "compound"}

    def __call__(self):
        return self.cls(self.left(), self.right())

    @classmethod
    def _from_object(cls, compound):
        return Compound(
            compound_type=compound.__class__.__name__,
            children=[
                Object.from_object(compound.left),
                Object.from_object(compound.right),
            ],
            cls=type(compound),
        )



from functools import cached_property
from typing import Tuple
from autofit.aggregator.search_output import SearchOutput


class Row:
    def __init__(
        self,
        result: SearchOutput,
        columns: list,
        number: int,
    ):
        """
        A row in the summary table corresponding to one search.

        Parameters
        ----------
        result
            The search output from the aggregator
        columns
            The columns to include in the summary. These are taken from samples_summary or latent_summary
        """
        self.result = result
        self._columns = columns
        self.number = number

    def _all_paths(self, path: Tuple[str, ...]):
        model = self.result.model
        return model.all_paths_for_prior(model.object_for_path(path))

    def _add_paths(self, kwargs):
        return {
            path: value
            for key, value in kwargs.items()
            for path in self._all_paths(
                key.split(".") if isinstance(key, str) else key,
            )
        }

    @property
    def median_pdf_sample_kwargs(self) -> dict:
        """
        The median_pdf_sample arguments for the search from the samples_summary and latent_summary.
        """
        samples_summary = self.result.value("samples_summary")
        kwargs = self._add_paths(samples_summary.median_pdf_sample.kwargs)

        latent_summary = self.result.value("latent.latent_summary")
        if latent_summary is not None:
            kwargs.update(latent_summary.median_pdf_sample.kwargs)

        return kwargs

    @property
    def max_likelihood_kwargs(self):
        """
        The median_pdf_sample arguments for the search from the samples_summary and latent_summary.
        """
        samples_summary = self.result.samples_summary
        kwargs = self._add_paths(samples_summary.median_pdf_sample.kwargs)

        latent_summary = self.result.value("latent.latent_summary")
        if latent_summary is not None:
            kwargs.update(latent_summary.max_log_likelihood_sample.kwargs)

        return kwargs

    @cached_property
    def values_at_sigma_1_kwargs(self) -> dict:
        """
        The values_at_sigma_1 arguments for the search from the samples_summary.
        """
        kwargs = self._add_paths(self.result.samples_summary.values_at_sigma_1)

        latent_summary = self.result.value("latent.latent_summary")
        if latent_summary is not None:
            kwargs.update(
                {
                    tuple(key.split(".")): value
                    for key, value in latent_summary.values_at_sigma_1.items()
                }
            )

        return kwargs

    @cached_property
    def values_at_sigma_3_kwargs(self) -> dict:
        """
        The values_at_sigma_3 arguments for the search from the samples_summary.
        """
        kwargs = self._add_paths(self.result.samples_summary.values_at_sigma_3)

        latent_summary = self.result.value("latent.latent_summary")
        if latent_summary is not None:
            kwargs.update(
                {
                    tuple(key.split(".")): value
                    for key, value in latent_summary.values_at_sigma_1.items()
                }
            )

        return kwargs

    def dict(self) -> dict:
        """
        The row as a dictionary including an id and one entry for each column.
        """
        row = {"id": self.result.id}
        for column in self._columns:
            value = column.value(self)
            if isinstance(value, dict):
                for key, value in value.items():
                    row[f"{column.name}_{key}" if key else column.name] = value
            else:
                row[column.name] = column.value(self)

        return row

    @property
    def column_names(self):
        """
        The names of the columns.
        """
        return list(self.dict().keys())



from typing import List, Union

from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.samples import Samples
from autofit.database import Prior
from autofit.non_linear.result import Result, Placeholder
from .job import JobResult
from .result import GridSearchResult


class ResultBuilder:
    def __init__(
        self,
        lists: List[List[float]],
        grid_priors: List[Prior],
        paths: List[AbstractPaths],
    ):
        """
        Builds GridSearchResults including all results so far computed
        and Placeholders where no result has yet been computed.

        Parameters
        ----------
        lists
            A list of lists of unit hypercube vectors describing the
            values that change during the grid search
        grid_priors
            Priors which are varied throughout the grid search
        """
        self.lists = lists
        self.grid_priors = grid_priors
        self._job_result_dict = {}
        self.paths = paths

    def __call__(self) -> GridSearchResult:
        """
        Generate a GridSearchResult with all results so far and placeholders
        where no result has been returned yet
        """
        return GridSearchResult(self.sample_summaries, self.lists, self.grid_priors)

    @property
    def results(self) -> List[Union[Result, Placeholder]]:
        """
        A list of results that have been returned with placeholders where no
        result has been returned in the grid-search order.
        """
        return [
            summary
            if isinstance(summary, Placeholder)
            else Result(
                samples_summary=summary,
                paths=paths,
            )
            for summary, paths in zip(self.sample_summaries, self.paths)
        ]

    @property
    def sample_summaries(self) -> List[Union[Samples, Placeholder]]:
        """
        A list of results that have been returned with placeholders where no
        result has been returned in the grid-search order.
        """
        samples = []
        for number in range(len(self.lists)):
            try:
                job_result = self._job_result_dict[number]
                samples.append(
                    job_result.result.samples_summary,
                )
            except KeyError:
                samples.append(Placeholder())
        return samples

    def add(self, job_result: JobResult):
        """
        Add a new result for a completed job.

        The number associated with the job result is used for
        tracking and ordering.

        Parameters
        ----------
        job_result
            A result of an optimisation for a single point in
            the grid.
        """
        self._job_result_dict[job_result.number] = job_result



import abc
import inspect
from sqlalchemy.orm import Mapped, DeclarativeMeta
from typing import List, Tuple, Any, Iterable, Union, ItemsView, Type

import numpy as np

from autoconf.class_path import get_class, get_class_path
from ..sqlalchemy_ import sa
from sqlalchemy.orm import declarative_base


class DeclarativeABCMeta(DeclarativeMeta, abc.ABCMeta):
    """A metaclass combining SQLAlchemy's DeclarativeMeta and ABCMeta."""


Base = declarative_base(metaclass=DeclarativeABCMeta)

_schema_version = 1


class Object(Base):
    __tablename__ = "object"

    type = sa.Column(sa.String)

    id = sa.Column(
        sa.Integer,
        primary_key=True,
        index=True,
    )

    parent_id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        index=True,
    )
    parent = sa.orm.relationship("Object", uselist=False, remote_side=[id])

    samples_for_id = sa.Column(sa.String, sa.ForeignKey("fit.id"))
    samples_for = sa.orm.relationship(
        "Fit",
        uselist=False,
        foreign_keys=[samples_for_id],
        back_populates="_samples",
    )

    latent_samples_for_id = sa.Column(sa.String, sa.ForeignKey("fit.id"))
    latent_samples_for = sa.orm.relationship(
        "Fit",
        uselist=False,
        foreign_keys=[latent_samples_for_id],
        back_populates="_latent_samples",
    )

    children: Mapped[List["Object"]] = sa.orm.relationship(
        "Object",
        uselist=True,
        back_populates="parent",
    )

    def __len__(self):
        return len(self.children)

    name = sa.Column(sa.String)

    __mapper_args__ = {"polymorphic_identity": "object", "polymorphic_on": type}

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name}>"

    # noinspection PyProtectedMember
    @classmethod
    def from_object(cls, source, name=""):
        """
        Create a database object for an object in a model.

        The specific database class used depends on the type of
        the object.

        If __getstate__ is defined for any instance then that
        dictionary is used in place of the __dict__ when
        serialising.

        Parameters
        ----------
        source
            A model
        name
            The name of the object wrt its parent

        Returns
        -------
        An instance of a concrete child of this class
        """

        from autofit.mapper.prior_model.prior_model import Model
        from autofit.mapper.prior.abstract import Prior
        from autofit.mapper.prior_model.collection import Collection

        from autofit.mapper.prior.arithmetic.compound import Compound

        if source is None or isinstance(
            source,
            (
                np.broadcast,
                abc.ABCMeta,
                np.ufunc,
            ),
        ):
            from .instance import NoneInstance

            instance = NoneInstance()
        elif isinstance(source, np.ndarray):
            from .array import Array

            # noinspection PyArgumentList
            instance = Array(array=source)
        elif isinstance(source, Model):
            from .prior import Model

            instance = Model._from_object(source)
        elif isinstance(source, Prior):
            from .prior import Prior

            instance = Prior._from_object(source)
        elif isinstance(source, (float, int)):
            from .instance import Value

            instance = Value._from_object(source)
        elif isinstance(source, (tuple, list)):
            from .instance import Collection

            instance = Collection._from_object(source)
        elif isinstance(source, dict):
            from .common import Dict

            instance = Dict._from_object(source)
        elif isinstance(source, Collection):
            from .prior import Collection

            instance = Collection._from_object(source)
        elif isinstance(source, str):
            from .instance import StringValue

            instance = StringValue._from_object(source)

        elif isinstance(source, Compound):
            from .compound import Compound

            instance = Compound._from_object(source)
        elif inspect.isfunction(source):
            from .function import Function

            instance = Function._from_object(source)
        else:
            from .instance import Instance

            instance = Instance._from_object(source)
        instance.name = name
        return instance

    @property
    def _constructor_args(self):
        return set(inspect.getfullargspec(self.cls).args[1:])

    def _make_instance(self) -> object:
        """
        Create the real instance for this object
        """
        try:
            return object.__new__(self.cls)
        except TypeError as e:
            raise TypeError(
                f"Could not instantiate {self.name} of type {self.cls}"
            ) from e

    def __call__(self):
        """
        Create the real instance for this object, with child
        attributes attached.

        If the instance implements __setstate__ then this is
        called with a dictionary of instantiated children.
        """
        instance = self._make_instance()
        child_instances = {child.name: child() for child in self.children}

        if hasattr(instance, "__setstate__"):
            instance.__setstate__(child_instances)
        else:
            for name, child in child_instances.items():
                setattr(instance, name, child)

        from autofit import ModelObject

        if isinstance(instance, ModelObject) and (
            not hasattr(instance, "id") or instance.id is None
        ):
            instance.id = instance.next_id()

        return instance

    def _add_children(
        self, items: Union[ItemsView[str, Any], Iterable[Tuple[str, Any]]]
    ):
        """
        Add database representations of child attributes

        Parameters
        ----------
        items
            Attributes such as floats or priors that are associated
            with the real object
        """
        for key, value in items:
            if isinstance(value, property) or key.startswith("__") or key == "dtype":
                continue
            child = Object.from_object(value, name=key)
            self.children.append(child)

    class_path = sa.Column(sa.String)

    @property
    def cls(self) -> Type[object]:
        """
        The class of the real object
        """
        return get_class(self.class_path)

    @cls.setter
    def cls(self, cls: type):
        self.class_path = get_class_path(cls)



import logging
from abc import ABC, abstractmethod
from sqlalchemy import text
from typing import Optional, List, Union, cast

from ..sqlalchemy_ import sa

from autofit.database import query as q
from autofit.database.aggregator.info import Info
from .scrape import Scraper
from autofit.database import model as m
from ..query.query import AbstractQuery, Attribute
from ..query.query.attribute import BestFitQuery

logger = logging.getLogger(__name__)


class NullPredicate(AbstractQuery):
    @property
    def fit_query(self) -> str:
        return "SELECT id FROM fit"

    def __and__(self, other):
        return other


class Query:
    """
    API for creating a query on the best fit instance
    """

    @staticmethod
    def for_name(name: str) -> q.Q:
        """
        Create a query for fits based on the name of a
        top level instance attribute

        Parameters
        ----------
        name
            The name of the attribute. e.g. galaxies

        Returns
        -------
        A query generating object
        """
        return q.Q(name)

    def __getattr__(self, name):
        return self.for_name(name)


class FitQuery(Query):
    """
    API for creating a query on the attributes of a fit,
    such as:
        name
        unique_tag
        path_prefix
        is_complete
        is_grid_search
    """

    @staticmethod
    def for_name(name: str) -> Union[AbstractQuery, Attribute]:
        """
        Create a query based on some attribute of the Fit.

        Parameters
        ----------
        name
            The name of an attribute of the Fit class

        Returns
        -------
        A query based on an attribute

        Examples
        --------
        aggregator.fit.name == 'example name'
        """
        if name not in m.fit_attributes:
            raise AttributeError(f"Fit has no attribute {name}")
        if m.fit_attributes[name].type.python_type == bool:
            return q.BA(name)
        return q.A(name)


class Reverse:
    def __init__(self, item):
        self.item = item

    @property
    def attribute(self):
        return self.item.attribute


class AbstractAggregator(ABC):
    @property
    @abstractmethod
    def fits(self) -> List[m.Fit]:
        pass

    def values(self, name: str, parser=lambda o: o) -> list:
        """
        Retrieve the value associated with each fit with the given
        parameter name

        Parameters
        ----------
        name
            The name of some pickle, such as 'samples'
        parser
            A function to parse the value

        Returns
        -------
        A list of objects, one for each fit
        """
        values = list()
        for fit in self:
            value = fit[name]
            if value is not None:
                values.append(parser(value))

        return values

    def child_values(self, name: str) -> List[list]:
        """
        Retrieve the value associated with each fit with the given
        parameter name

        Parameters
        ----------
        name
            The name of some pickle, such as 'samples'

        Returns
        -------
        A list of objects, one for each fit
        """
        return [fit.child_values(name) for fit in self]

    def __iter__(self):
        return iter(self.fits)

    def __len__(self):
        return len(self.fits)

    def __eq__(self, other):
        if isinstance(other, list):
            return self.fits == other
        return super().__eq__(other)


class Aggregator(AbstractAggregator):
    def __init__(
        self,
        session: sa.orm.Session,
        filename: Optional[str] = None,
        predicate: AbstractQuery = NullPredicate(),
        offset=0,
        limit=None,
        order_bys=None,
        top_level_only=True,
    ):
        """
        Query results from an intermediary SQLite database.

        Results can be scraped from a directory structure and stored in the database.

        Parameters
        ----------
        session
            A session for communicating with the database.
        filename
            The path to the database file. If None, the database is in memory.
        predicate
            A predicate to filter the results by.
        offset
            The number of results to skip
        limit
            The maximum number of results to return
        order_bys
            A list of attributes to order the results by
        top_level_only
            If True, only return the top level fits
        """
        self.session = session
        self.filename = filename
        self._fits = None
        self._predicate = predicate
        self._offset = offset
        self._limit = limit
        self._order_bys = order_bys or list()
        self._top_level_only = top_level_only

    def order_by(self, item: Attribute, reverse=False) -> "Aggregator":
        """
        Order the results by a given attribute of the search. Can be applied
        multiple times with the first application taking precedence.

        Parameters
        ----------
        item
            An attribute of the search
        reverse
            If True reverse the results

        Returns
        -------
        An aggregator with ordering applied

        Examples
        --------
        aggregator = aggregator.order_by(
            aggregator.search.unique_tag
        )
        """
        if reverse:
            item = Reverse(item)
        return self._new_with(order_bys=self._order_bys + [item])

    @property
    def search(self) -> FitQuery:
        """
        An object facilitating queries on fit attributes such as:
            name
            unique_tag
            path_prefix
            is_complete
            is_grid_search
        """
        return FitQuery()

    @property
    def info(self):
        """
        Query info associated with the fit in the info dictionary
        """
        return q.AnonymousInfo()

    @property
    def fits(self) -> List[m.Fit]:
        """
        Lazily query the database for a list of Fit objects that
        match the aggregator's predicate.
        """
        if self._fits is None:
            self._fits = self._fits_for_query(self._predicate.fit_query)
        return self._fits

    def map(self, func):
        for fit in self.fits:
            yield func(fit)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.filename} {len(self)}>"

    @property
    def model(self) -> Query:
        """
        Facilitates query construction. If the Fit class has an
        attribute with the given name then a predicate is generated
        based on that attribute. Otherwise the query is assumed to
        apply to the best fit instance.

        Returns
        -------
        A query
        """
        return Query()

    def __call__(self, predicate) -> "Aggregator":
        """
        Concise query syntax
        """
        return self.query(predicate)

    def query(self, predicate: AbstractQuery) -> "Aggregator":
        # noinspection PyUnresolvedReferences
        """
        Apply a query on the model.

        Parameters
        ----------
        predicate
            A predicate constructed to express which models should be included.

        Returns
        -------
        A list of objects that match the predicate

        Examples
        --------
        >>>
        >>> aggregator = Aggregator.from_database(
        >>>     "my_database.sqlite"
        >>> )
        >>>
        >>> lens = aggregator.galaxies.lens
        >>>
        >>> aggregator.filter((lens.bulge == SersicCore) & (lens.disk == Sersic))
        >>> aggregator.filter((lens.bulge == SersicCore) | (lens.disk == Sersic))
        """
        return self._new_with(predicate=self._predicate & predicate)

    def _new_with(
        self,
        type_=None,
        **kwargs,
    ) -> "Aggregator":
        """
        Create a new instance with the same attribute values except
        for those overridden by kwargs

        Parameters
        ----------
        type_
            The type of the new instance (defaults to Aggregator)
        kwargs
            Names and values of attributes to override

        Returns
        -------
        A new Aggregator with the same attributes except where they
        have been overridden
        """
        kwargs = {
            "session": self.session,
            "filename": self.filename,
            "predicate": self._predicate,
            "order_bys": self._order_bys,
            "top_level_only": self._top_level_only,
            **kwargs,
        }
        type_ = type_ or type(self)
        return type_(**kwargs)

    def __getitem__(self, item):
        offset = self._offset
        limit = self._limit
        if isinstance(item, int):
            return self.fits[item]
        elif isinstance(item, slice):
            if item.start is not None:
                if item.start >= 0:
                    offset += item.start
                else:
                    offset = len(self) + item.start
            if item.stop is not None:
                if item.stop >= 0:
                    limit = len(self) - item.stop - offset
                else:
                    limit = len(self) + item.stop
        return self._new_with(offset=offset, limit=limit)

    def _fits_for_query(self, query: str) -> List[m.Fit]:
        """
        Execute a raw SQL query and return a Fit object
        for each Fit id returned by the query

        Parameters
        ----------
        query
            A SQL query that selects ids from the fit table

        Returns
        -------
        A list of fit objects, one for each id returned by the
        query
        """
        logger.debug(f"Executing query: {query}")
        fit_ids = {row[0] for row in self.session.execute(text(query))}

        logger.info(f"{len(fit_ids)} fit(s) found matching query")
        query = self.session.query(m.Fit).filter(m.Fit.id.in_(fit_ids))
        for order_by in self._order_bys:
            attribute = getattr(m.Fit, order_by.attribute)

            if isinstance(order_by, Reverse):
                attribute = sa.desc(attribute)
            query = query.order_by(attribute)

        fits = query.offset(self._offset).limit(self._limit).all()
        if self._top_level_only:
            return [fit for fit in fits if fit.parent is None]
        return fits

    def add_directory(
        self,
        directory: str,
        auto_commit=True,
        reference: Optional[dict] = None,
        completed_only: bool = False,
    ):
        """
        Recursively search a directory for autofit results
        and add them to this database.

        Any pickles found in the pickles file are implicitly added
        to the fit object.

        Warnings
        --------
        If a directory is added twice then that will result in
        duplicate entries in the database.

        Parameters
        ----------
        auto_commit
            If True the session is committed writing the new objects
            to the database
        directory
            A directory containing autofit results embedded in a
            file structure
        reference
            A dictionary mapping the names of objects in the model
            to their class path.
        completed_only
            If true only searches that have completed are added
        """
        scraper = Scraper(
            directory,
            self.session,
            reference=reference,
            completed_only=completed_only,
        )
        scraper.scrape()

        if auto_commit:
            self.session.commit()
            Info(self.session).write()

    @classmethod
    def from_database(
        cls,
        filename: str,
        completed_only: bool = False,
        top_level_only: bool = True,
    ) -> "Aggregator":
        """
        Create an instance from a sqlite database file.

        If no file exists then one is created with the schema of the database.

        Parameters
        ----------
        completed_only
            If True only completed fits are returned
        filename
            The name of the database file.
        top_level_only
            If True only top level fits are returned

        Returns
        -------
        An aggregator connected to the database specified by the file.
        """
        from autofit.database import open_database

        session = open_database(str(filename))
        aggregator = Aggregator(session, filename, top_level_only=top_level_only)
        if completed_only:
            return aggregator(aggregator.search.is_complete)
        return aggregator

    def grid_searches(self) -> "GridSearchAggregator":
        """
        Filter to only grid searches and return an aggregator
        with grid search specific functionality.

        Grid searches are initially implicitly ordered by their id
        """
        return cast(
            GridSearchAggregator,
            self._new_with(
                type_=GridSearchAggregator,
                predicate=self._predicate & self.search.is_grid_search,
                order_bys=[Attribute("id")],
                top_level_only=False,
            ),
        )


class GridSearchAggregator(Aggregator):
    def __init__(
        self,
        session: sa.orm.Session,
        filename: Optional[str] = None,
        predicate: AbstractQuery = NullPredicate(),
        offset=0,
        limit=None,
        order_bys=None,
        top_level_only=False,
    ):
        super().__init__(
            session=session,
            filename=filename,
            predicate=predicate,
            offset=offset,
            limit=limit,
            order_bys=order_bys,
            top_level_only=top_level_only,
        )

    def best_fits(self) -> "GridSearchAggregator":
        """
        The best fit from each of the grid searches

        Best fits are initially implicitly ordered by their parent id
        """
        return self._new_with(
            predicate=BestFitQuery(self._predicate), order_bys=[Attribute("parent_id")]
        )

    def children(self) -> "GridSearchAggregator":
        """
        An aggregator comprising the children of the fits encapsulated
        by this aggregator. This is used to query children in a grid search.

        Children are initially implicitly ordered by their parent id
        """
        return self._new_with(
            predicate=q.ChildQuery(self._predicate), order_bys=[Attribute("parent_id")]
        )

    def cell_number(self, number: int) -> "CellAggregator":
        """
        Create an aggregator for accessing all values for child fits
        with a given index, ordered by parameter values.

        Parameters
        ----------
        number
            The number of the fit in the grid search

        Returns
        -------
        An aggregator comprising fits for a given cell for each grid search
        """
        return CellAggregator(number, self)


class CellAggregator(AbstractAggregator):
    def __init__(self, number: int, aggregator: GridSearchAggregator):
        """
        Aggregator for accessing data for a specific fit number in each
        grid search.

        Parameters
        ----------
        number
            The number of the fit
        aggregator
            An aggregator comprising 0 or more grid searches
        """
        self.number = number
        self.aggregator = aggregator
        self._fits = None

    @property
    def fits(self) -> List[m.Fit]:
        """
        Retrieve one fit for each grid search matching the number of
        the cell.
        """
        if self._fits is None:
            self._fits = list()
            for fit in self.aggregator:
                self._fits.append(
                    sorted(
                        fit.children,
                        key=lambda f: f.model.order_no if f.model is not None else 0,
                    )[self.number]
                )
        return self._fits



from copy import copy
from itertools import count
from typing import Callable, Optional

from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.parallel import AbstractJob, AbstractJobResult
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.result import Result


class JobResult(AbstractJobResult):
    def __init__(self, number: int, result: Result, perturb_result: Result):
        """
        The result of a single sensitivity comparison

        Parameters
        ----------
        result
        perturb_result
        """
        super().__init__(number)
        self.result = result
        self.perturb_result = perturb_result

    @property
    def log_evidence_increase(self) -> Optional[float]:
        """
        Returns a tuple of the log evidence of the base model, the perturbed model and the difference between them.

        This is used to ouptut the sensitivity mapping results to .csv files.

        If the log evidence is not available, a tuple containing 3 None's is returned.
        """

        if hasattr(self.result.samples, "log_evidence"):
            if self.result.samples.log_evidence is not None and self.perturb_result.samples.log_evidence is not None:
                return float(
                    self.perturb_result.samples.log_evidence
                    - self.result.samples.log_evidence
                )

    @property
    def log_likelihood_increase(self) -> Optional[float]:
        """
        Returns a tuple of the log likelihood of the base model, the perturbed model and the difference between them.

        This is used to ouptut the sensitivity mapping results to .csv files.
        """

        return float(self.perturb_result.log_likelihood - self.result.log_likelihood)


class MaskedJobResult(AbstractJobResult):
    """
    A placeholder result for a job that has been masked out.
    """

    def __init__(self, number, model):
        super().__init__(number)
        self.model = model

    @property
    def result(self):
        return self

    @property
    def perturb_result(self):
        return self

    def __getattr__(self, item):
        return None

    @property
    def samples_summary(self):
        return self

    @property
    def log_evidence(self):
        return 0.0

    @property
    def log_likelihood(self):
        return 0.0


class Job(AbstractJob):
    _number = count()

    def __init__(
        self,
        model: AbstractPriorModel,
        simulate_cls: Callable,
        perturb_model: AbstractPriorModel,
        simulate_instance: ModelInstance,
        base_instance: ModelInstance,
        base_fit_cls: Callable,
        perturb_fit_cls: Callable,
        paths: AbstractPaths,
        number: int,
    ):
        """
        Job to run non-linear searches comparing how well a model and a model with a perturbation fit the image.

        Parameters
        ----------
        model
            A base model that fits the image without a perturbation
        perturb_model
            A model of the perturbation which has been added to the underlying image
        base_fit_cls
            A class which defines the function which fits the base model to each simulated dataset of the sensitivity
            map.
        perturb_fit_cls
            A class which defines the function which fits the perturbed model to each simulated dataset of the
            sensitivity map.
        paths
            The paths defining the output directory structure of the sensitivity mapping.
        """
        super().__init__(number=number)

        self.model = model
        self.simulate_cls = simulate_cls
        self.perturb_model = perturb_model
        self.simulate_instance = simulate_instance
        self.base_instance = base_instance
        self.base_fit_cls = base_fit_cls
        self.perturb_fit_cls = perturb_fit_cls
        self.paths = paths

    @property
    def base_paths(self):
        return self.paths.for_sub_analysis("[base]")

    @property
    def perturb_paths(self):
        return self.paths.for_sub_analysis("[perturb]")

    @property
    def is_complete(self) -> bool:
        """
        Returns True if the job has been completed, False otherwise.
        """
        return (self.base_paths.is_complete and self.perturb_paths.is_complete) or (
            (self.paths.output_path / "[base].zip").exists()
            and (self.paths.output_path / "[perturb].zip").exists()
        )

    def perform(self) -> JobResult:
        """
        - Create one model with a perturbation and another without
        - Fit each model against the perturbed image

        Returns
        -------
        An object comprising the results of the two fits
        """

        if self.is_complete:
            dataset = None
        else:
            dataset = self.simulate_cls(
                instance=self.simulate_instance,
                simulate_path=self.paths.image_path.with_name("simulate"),
            )

        result = self.base_fit_cls(
            model=self.model,
            dataset=dataset,
            paths=self.base_paths,
            instance=self.simulate_instance,
        )

        perturb_model = copy(self.model)
        perturb_model.perturb = self.perturb_model

        perturb_result = self.perturb_fit_cls(
            model=perturb_model,
            dataset=dataset,
            paths=self.perturb_paths,
            instance=self.simulate_instance,
        )

        return JobResult(
            number=self.number,
            result=result,
            perturb_result=perturb_result,
        )



from __future__ import annotations
import logging
from abc import ABC, abstractmethod
import numpy as np
from typing import TYPE_CHECKING, Optional
import warnings

if TYPE_CHECKING:
    from autofit.non_linear.analysis.analysis import Analysis

from autofit import exc
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.samples import Samples
from autofit.non_linear.samples.summary import SamplesSummary
from autofit.text import text_util


class Placeholder:
    def __getattr__(self, item):
        """
        Placeholders return None to represent the missing result's value
        """
        return None

    def __getstate__(self):
        return {}

    def __setstate__(self, state):
        pass

    def __gt__(self, other):
        return False

    def __lt__(self, other):
        return True

    @property
    def samples(self):
        return self

    @property
    def log_likelihood(self):
        return -np.inf

    def summary(self):
        return self


class AbstractResult(ABC):
    """
    @DynamicAttrs
    """

    def __init__(self, samples_summary, paths):
        """
        Abstract result of a non-linear search.

        Parameters
        ----------
        samples_summary
            A summary of the most important samples of the non-linear search (e.g. maximum log likelihood, median PDF).
        paths
            The paths to the results of the search.
        """

        self._samples_summary = samples_summary
        self.paths = paths

    @property
    def samples_summary(self):
        return self._samples_summary

    @property
    @abstractmethod
    def samples(self):
        pass

    @property
    @abstractmethod
    def model(self):
        pass

    @property
    def info(self) -> str:
        return text_util.result_info_from(
            samples=self.samples,
        )

    def __gt__(self, other):
        """
        Results are sorted by their associated log_likelihood.

        Placeholders are always low.
        """
        if isinstance(other, Placeholder):
            return True
        return self.log_likelihood > other.log_likelihood

    def __lt__(self, other):
        """
        Results are sorted by their associated log_likelihood.

        Placeholders are always low.
        """
        if isinstance(other, Placeholder):
            return False
        return self.log_likelihood < other.log_likelihood

    @property
    def log_likelihood(self):
        return self.samples_summary.max_log_likelihood_sample.log_likelihood

    @property
    def instance(self):
        try:
            return self.samples_summary.instance
        except AttributeError as e:
            logging.warning(e)
            return None

    @property
    def max_log_likelihood_instance(self):
        return self.instance

    @property
    def model_centred(self) -> AbstractPriorModel:
        """
        Returns a model where every free parameter is a `GaussianPrior` with `mean` the previous result's
        inferred maximum log likelihood parameter values and `sigma` the input absolute value `a`.

        For example, a previous result may infer a parameter to have a maximum log likelihood value of 2.

        If this result is used for search chaining, `model_centred_absolute(a=0.1)` will assign this free parameter
        `GaussianPrior(mean=2.0, sigma=0.1)` in the new model, where `sigma` is linked to the input `a`.

        Parameters
        ----------
        a
            The absolute width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this search and creating priors with the defined absolute
        width.
        """
        return self.samples_summary.model_centred

    def model_centred_absolute(self, a: float) -> AbstractPriorModel:
        """
        Returns a model where every free parameter is a `GaussianPrior` with `mean` the previous result's
        inferred maximum log likelihood parameter values and `sigma` the input absolute value `a`.

        For example, a previous result may infer a parameter to have a maximum log likelihood value of 2.

        If this result is used for search chaining, `model_centred_absolute(a=0.1)` will assign this free parameter
        `GaussianPrior(mean=2.0, sigma=0.1)` in the new model, where `sigma` is linked to the input `a`.

        Parameters
        ----------
        a
            The absolute width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this search and creating priors with the defined absolute
        width.
        """
        return self.samples_summary.model_centred_absolute(a)

    def model_centred_relative(self, r: float) -> AbstractPriorModel:
        """
        Returns a model where every free parameter is a `GaussianPrior` with `mean` the previous result's
        inferred maximum log likelihood parameter values and `sigma` a relative value from the result `r`.

        For example, a previous result may infer a parameter to have a maximum log likelihood value of 2 and
        an error at the input `sigma` of 0.5.

        If this result is used for search chaining, `model_centred_relative(r=0.1)` will assign this free parameter
        `GaussianPrior(mean=2.0, sigma=0.5*0.1)` in the new model, where `sigma` is the inferred error times `r`.

        Parameters
        ----------
        r
            The relative width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this search and creating priors with the defined relative
        width.
        """
        return self.samples_summary.model_centred_relative(r)

    def model_centred_max_lh_bounded(self, b: float) -> AbstractPriorModel:
        """
        Returns a model where every free parameter is a `UniformPrior` with `lower_limit` and `upper_limit` the previous
        result's inferred maximum log likelihood parameter values minus and plus the bound `b`.

        For example, a previous result may infer a parameter to have a maximum log likelihood value of 2.

        If this result is used for search chaining, `model_bound(b=0.1)` will assign this free parameter
        `UniformPrior(lower_limit=1.9, upper_limit=2.1)` in the new model.

        Parameters
        ----------
        b
            The size of the bounds of the uniform prior

        Returns
        -------
        A model mapper created by taking results from this search and creating priors with the defined bounded
        uniform priors.
        """
        return self.samples_summary.model_centred_max_lh_bounded(b)


class Result(AbstractResult):
    def __init__(
        self,
        samples_summary: SamplesSummary,
        paths: Optional[AbstractPaths] = None,
        samples: Optional[Samples] = None,
        search_internal: Optional[object] = None,
        analysis: Optional[Analysis] = None,
    ):
        """
        The result of a non-linear search.

        The default behaviour is for all key results to be in the `samples_summary` attribute, which is a concise
        summary of the results of the non-linear search. The reasons for this to be the main attribute are:

        - It is concise and therefore has minimal I/O overhead, which is important because when runs are resumed
        the results are loaded often, which can become very slow for large results via a `samples.csv`.

        - The `output.yaml` config files can be used to disable the output of the `samples.csv` file
        and `search_internal.dill` files. This means in order for results to be loaded in a way that allows a run to
        resume, the `samples_summary` must contain all results necessary to resume the run.

        For this reason, the `samples` and `search_internal` attributes are optional. On the first run of a model-fit,
        they will always contain values as they are passed in via memory from the results of the search. However, if
        a run is resumed they are no longer available in memory, and they will only be available if their corresponding
        `samples.csv` and `search_internal.dill` files are output on disk and available to load.

        This object includes:

        - The `samples_summary` attribute, which is a summary of the results of the non-linear search.

        - The `paths` attribute, which contains the path structure to the results of the search on the hard-disk and
        is used to load the samples and search internal attributes if they are required and not available in memory.

        - The samples of the non-linear search (E.g. MCMC chains, nested sampling samples) which are used to compute
        the maximum likelihood model, posteriors and other properties.

        - The non-linear search used to perform the model fit in its internal format (e.g. the Dynesty sampler used
        by dynesty itself as opposed to PyAutoFit abstract classes).

        Parameters
        ----------
        samples_summary
            A summary of the most important samples of the non-linear search (e.g. maximum log likelihood, median PDF).
        paths
            The paths to the results of the search, used to load the samples and search internal attributes if they are
            required and not available in memory.
        samples
            The samples of the non-linear search, for example the MCMC chains.
        search_internal
            The non-linear search used to perform the model fit in its internal format.
        analysis
            The `Analysis` object that was used to perform the model-fit from which this result is inferred.
        """
        super().__init__(samples_summary=samples_summary, paths=paths)

        self._samples = samples
        self._search_internal = search_internal

        self.analysis = analysis

        self.__model = None

        self.child_results = None

    def dict(self) -> dict:
        """
        Human-readable dictionary representation of the results
        """
        return {
            "max_log_likelihood": self.samples_summary.max_log_likelihood_sample.model_dict(),
            "median pdf": self.samples_summary.median_pdf_sample.model_dict(),
        }

    @property
    def samples(self) -> Optional[Samples]:
        """
        Returns the samples of the non-linear search, for example the MCMC chains or nested sampling samples.

        When a model-fit is run the first time, the samples are passed into the result via memory and therefore
        always available.

        However, if a model-fit is resumed the samples are not available in memory and they only way to load them is
        via the `samples.csv` file output on the hard-disk. This property handles the loading of the samples from
        the `samples.csv` file if they are not available in memory.

        Returns
        -------
        The samples of the non-linear search.
        """

        if self._samples is not None:
            return self._samples

        try:
            return self.paths.samples
        except FileNotFoundError:
            return None

    @property
    def search_internal(self):
        """
        Returns the non-linear search used to perform the model fit in its internal sampler format.

        When a model-fit is run the first time, the search internal is passed into the result via memory and therefore
        always available.

        However, if a model-fit is resumed the search internal is not available in memory and they only way to load
        it is via the `search_internal.dill` file output on the hard-disk. This property handles the loading of
        the search internal from the `search_internal.dill` file if it is not available in memory.

        Returns
        -------
        The non-linear search used to perform the model fit in its internal sampler format.
        """
        if self._search_internal is not None:
            return self._search_internal

        try:
            return self.paths.load_search_internal()
        except FileNotFoundError:
            pass

    @property
    def projected_model(self) -> AbstractPriorModel:
        """
        Create a new model with the same structure as the previous model,
        replacing each prior with a new prior created by calculating sufficient
        statistics from samples and corresponding weights for that prior.
        """
        weights = self.samples.weight_list
        arguments = {
            prior: prior.project(
                samples=np.array(self.samples.values_for_path(path)),
                weights=weights,
            )
            for path, prior in self.samples.model.path_priors_tuples
        }
        return self.samples.model.mapper_from_prior_arguments(arguments)

    @property
    def model(self):
        if self.__model is None:
            self.__model = self.samples_summary.model.mapper_via_defaults_from()

        return self.__model

    @model.setter
    def model(self, model):
        self.__model = model

    def __str__(self):
        return "Analysis Result:\n{}".format(
            "\n".join(
                ["{}: {}".format(key, value) for key, value in self.__dict__.items()]
            )
        )

    def __getitem__(self, item):
        return self.child_results[item]

    def __iter__(self):
        return iter(self.child_results)

    def __len__(self):
        return len(self.child_results)


class ResultsCollection:
    def __init__(self, result_list=None):
        """
        A collection of results from previous searches. Results can be obtained using an index or the name of the search
        from whence they came.
        """
        self.__result_list = []
        self.__result_dict = {}

        if result_list is not None:
            for result in result_list:
                self.add(name="", result=result)

    def copy(self):
        collection = ResultsCollection()
        collection.__result_dict = self.__result_dict
        collection.__result_list = self.__result_list
        return collection

    @property
    def reversed(self):
        return reversed(self.__result_list)

    @property
    def last(self):
        """
        The result of the last search
        """
        if len(self.__result_list) > 0:
            return self.__result_list[-1]
        return None

    @property
    def first(self):
        """
        The result of the first search
        """
        if len(self.__result_list) > 0:
            return self.__result_list[0]
        return None

    def add(self, name, result):
        """
        Add the result of a search.

        Parameters
        ----------
        name: str
            The name of the search
        result
            The result of that search
        """
        try:
            self.__result_list[self.__result_list.index(result)] = result
        except ValueError:
            self.__result_list.append(result)
        self.__result_dict[name] = result

    def __getitem__(self, item):
        """
        Get the result of a previous search by index

        Parameters
        ----------
        item: int
            The index of the result

        Returns
        -------
        result: Result
            The result of a previous search
        """
        return self.__result_list[item]

    def __len__(self):
        return len(self.__result_list)

    def from_name(self, name):
        """
        Returns the result of a previous search by its name

        Parameters
        ----------
        name: str
            The name of a previous search

        Returns
        -------
        result: Result
            The result of that search

        Raises
        ------
        exc.PipelineException
            If no search with the expected result is found
        """
        try:
            return self.__result_dict[name]
        except KeyError:
            raise exc.PipelineException(
                "No previous search named {} found in results ({})".format(
                    name, ", ".join(self.__result_dict.keys())
                )
            )

    def __contains__(self, item):
        return item in self.__result_dict



from .abstract import AbstractInterpolator


class SplineInterpolator(AbstractInterpolator):
    """
    Interpolate data with a piecewise cubic polynomial which is twice continuously differentiable
    """

    @staticmethod
    def _relationship(x, y):

        from scipy.interpolate import CubicSpline

        return CubicSpline(x, y)



from abc import ABC, abstractmethod
from typing import Any, List, Optional


class Fit(ABC):
    """
    Common interface for objects representing the result of a "fit" or "search."
    This interface is intentionally minimal, representing the smallest
    set of attributes and methods that are conceptually shared by both
    a database-backed Fit and a file-backed SearchOutput.
    """

    children: List["Fit"]

    @property
    @abstractmethod
    def id(self) -> str:
        """
        A unique identifier for the fit/search output.
        """

    @property
    @abstractmethod
    def name(self) -> Optional[str]:
        """
        The name of the fit/search output.
        """

    @property
    @abstractmethod
    def unique_tag(self) -> Optional[str]:
        """
        A unique tag for differentiating this fit/search output from others.
        """

    @property
    @abstractmethod
    def path_prefix(self) -> Optional[str]:
        """
        A path prefix or directory path used by the search or fit.
        """

    @property
    @abstractmethod
    def is_complete(self) -> bool:
        """
        True if the fit/search completed successfully, False otherwise.
        """

    @property
    @abstractmethod
    def model(self) -> Any:
        """
        The model that was fit (database-backed or loaded from disk).
        """

    @property
    @abstractmethod
    def instance(self) -> Any:
        """
        The maximum-likelihood instance (database-backed or loaded from disk).
        """

    @property
    @abstractmethod
    def samples(self) -> Any:
        """
        The samples resulting from the fit/search.
        """

    @property
    @abstractmethod
    def latent_samples(self) -> Any:
        """
        An alternative set of samples (latent samples) if they exist, else None.
        """

    @abstractmethod
    def child_values(self, name: str) -> List[Any]:
        """
        Get the values of a given key for all children, if the concept of children applies.
        """

    @property
    @abstractmethod
    def children(self) -> List[Any]:
        """
        Return a list of child fits/searches, if the concept of children applies.
        """

    @abstractmethod
    def value(self, name: str) -> Any:
        """
        Retrieve a stored object by its name (Pickle, JSON, array, etc.).
        Returns None if no value with that name is found.
        """



from autofit.mapper.model import ModelInstance
from autofit.mapper.model_mapper import ModelMapper
from autofit.non_linear.result import Result

from autofit.non_linear.mock.mock_samples_summary import MockSamplesSummary
from autofit.non_linear.mock.mock_samples import MockSamples


class MockResult(Result):
    def __init__(
        self,
        samples_summary: MockSamplesSummary = None,
        paths=None,
        samples=None,
        instance=None,
        analysis=None,
        search=None,
        model=None,
    ):
        super().__init__(
            samples_summary=samples_summary
            or MockSamplesSummary(
                model=model or ModelMapper(),
            ),
            paths=paths,
            samples=samples,
            search_internal=None,
        )

        self._instance = instance or ModelInstance()
        self._samples = samples or MockSamples(
            model=model
            or ModelMapper()
        )

        self.prior_means = None
        self.analysis = analysis
        self.search = search
        self.model = model

    @property
    def model_centred(self):
        try:
            return self.samples_summary.model_centred
        except AttributeError:
            return self.model

    def model_centred_absolute(self, a):
        try:
            return self.samples_summary.model_centred_absolute(a)
        except AttributeError:
            return self.model

    def model_centred_relative(self, r):
        try:
            return self.samples_summary.model_centred_relative(r)
        except AttributeError:
            return self.model

    @property
    def last(self):
        return self


class MockResultGrid(Result):
    def __init__(self, log_likelihood):
        # noinspection PyTypeChecker
        super().__init__(None, None)
        self._log_likelihood = log_likelihood
        self.model = log_likelihood

    @property
    def log_likelihood(self):
        return self._log_likelihood

    @property
    def best_model(self):
        return self.model
